import os
import sys
import copy
import inspect
import contextlib
import collections
import logging
import json

import django.http.request

import debian.debian_support

import mini_buildd.misc

from mini_buildd.models.msglog import MsgLog
LOG = logging.getLogger(__name__)


class Argument():
    def __init__(self, id_list, doc="Undocumented", default=None, choices=None):
        """
        :param id_list: List like '['--with-rollbacks', '-R']' for option or '['distributions']' for positionals; 1st entry always denotes the id.

        >>> Argument(["--long-id", "-s"]).identity
        'long_id'
        >>> Argument(["posi-tional"]).identity
        'posi_tional'
        """
        self.id_list = id_list
        self.doc = doc
        self.default = default
        self.raw_value = self._v2r(default) if default is not None else None
        self.given = False
        self.argparse_is_positional = not id_list[0].startswith("--")

        # identity: 1st of id_list with leading '--' removed and hyphens turned to snake case
        self.identity = mini_buildd.misc.Snake(id_list[0][0 if self.argparse_is_positional else 2:]).from_kebab()

        # kvsargs for argparse
        self.argparse_kvsargs = {}
        self.argparse_kvsargs["help"] = doc
        if default is not None:
            self.argparse_kvsargs["default"] = default

        self.choices = [] if choices is None else choices

    def _r2v(self):
        """Raw to value. Pre: self.raw_value is not None."""
        return self.raw_value

    @classmethod
    def _v2r(cls, value):
        """Value to raw."""
        return str(value)

    def set(self, value):
        if isinstance(value, str):
            self.raw_value = value
        else:
            self.raw_value = self._v2r(value)
        self.given = True

    @property
    def value(self):
        """Get value, including convenience transformations."""
        return self._r2v() if self.raw_value is not None else None

    # do we really need that?
    def false2none(self):
        return self.raw_value if self.raw_value else None

    def update(self):
        """Overwrite to implement custom run-time update."""


class TextArgument(Argument):
    TYPE = "text"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.argparse_kvsargs["action"] = "store"


class URLArgument(TextArgument):
    TYPE = "url"


class TextareaArgument(TextArgument):
    TYPE = "textarea"


class NumberArgument(TextArgument):
    TYPE = "number"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.argparse_kvsargs["type"] = int

    def _r2v(self):
        return int(self.raw_value)


class SelectArgument(Argument):
    TYPE = "select"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.choices:
            self.argparse_kvsargs["choices"] = self.choices


class BoolArgument(SelectArgument):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, choices=[str(True), str(False)], **kwargs)

    def _r2v(self):
        return self.raw_value == str(True)


class MultiSelectArgument(TextArgument):
    TYPE = "multiselect"
    SEPARATOR = ","

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.argparse_kvsargs["nargs"] = "+"
        self.argparse_kvsargs["help"] += " (multiple items: use repeatedly)"
        self.doc += " (multiple items: comma-separate)"

        if self.choices:
            self.argparse_kvsargs["choices"] = self.choices

    def _r2v(self):
        return self.raw_value.split(self.SEPARATOR) if self.raw_value else []

    def _v2r(self, value):
        return self.SEPARATOR.join(value)


class RepositoryRegex(TextArgument):
    def __init__(self, ident):
        super().__init__(ident, default=".*", doc="repository name regex")

    def update(self):
        self.choices = [r.identity for r in mini_buildd.Daemon().get_active_repositories()]


class RepositoryPattern(TextArgument):
    def __init__(self, ident):
        super().__init__(ident, default=".*", doc="repository name regex")

    def update(self):
        RepositoryRegex.update(self)


class Distribution(TextArgument):
    def __init__(self, ident):
        super().__init__(ident, doc="distribution name (<codename>-<id>-<suite>[-rollback<n>])")

    def update(self, **kwargs):
        self.choices = []
        for r in mini_buildd.Daemon().get_active_repositories():
            self.choices += r.mbd_distribution_strings(**kwargs)


class Distributions(MultiSelectArgument):
    def __init__(self, ident):
        super().__init__(ident, doc="list of distributions")

    def update(self):
        Distribution.update(self)


class KeyringDistributions(Distributions):
    def update(self):
        Distribution.update(self, build_keyring_package=True)
        self.doc += "\nDefaults to all 'build_keyring_package distributions' if none given."
        if not self.given:
            self.set(self.choices)


class UploadableDistributions(Distributions):
    def update(self):
        Distribution.update(self, uploadable=True)


class AutoPortDistributions(UploadableDistributions):
    def update(self):
        super().update()
        if not self.given:
            self.set([])


class TestDistributions(UploadableDistributions):
    def update(self):
        super().update()
        # Default layout has two (snapshot and experimental) suites flagged as experimental.
        # So we go here for the string "experimental" (not the flag) to avoid double testing in the standard case.
        self.doc += "\nBuilds for all 'experimental' distributions if none given."
        if not self.given:
            self.set([d for d in self.choices if d.endswith("experimental")])


class FindDistributions(Distributions):
    def update(self):
        super().update()
        if not self.given:
            self.set(self.choices)


class Codename(TextArgument):
    def __init__(self, ident):
        super().__init__(ident, doc="codename (base distribution) to get apt lines for")

    def update(self):
        self.choices = mini_buildd.Daemon().get_active_codenames()


class Suite(TextArgument):
    def __init__(self, ident):
        super().__init__(ident, default=".*", doc="suite name regex")

    def update(self):
        self.choices = [s.name for s in mini_buildd.Daemon().get_suites()]


class Source(TextArgument):
    def __init__(self, ident):
        super().__init__(ident, doc="source package name")

    def update(self):
        self.choices = mini_buildd.Daemon().events.get_last_sources()


class PatternSource(TextArgument):
    def __init__(self, ident):
        super().__init__(ident, default="*", doc="limit packages by name (glob pattern)")

    def update(self):
        Source.update(self)


class SourceVersion(TextArgument):
    def __init__(self, ident):
        super().__init__(ident, default="", doc=("limit command to that version. Use it for the rare case of\n"
                                                 "multiple version of the same package in one distribution (in\n"
                                                 "different components), or just as safeguard\n"))


class UploadOptions(TextArgument):
    def __init__(self, ident):
        from mini_buildd.changes import Upload
        super().__init__(ident,
                         default="lintian-mode=ignore",
                         doc=("list of upload options, separated by '|':\n"
                              "\n"
                              f"{Upload.Options.usage()}"
                              ))


class Command():
    # Auth
    NONE = 0
    LOGIN = 1
    STAFF = 2
    ADMIN = 3
    AUTH = NONE
    AUTH_STRINGS = {NONE: "anonymous",
                    LOGIN: "any user login",
                    STAFF: "staff user login",
                    ADMIN: "super user login"}

    CONFIRM = False
    NEEDS_RUNNING_DAEMON = False
    ARGUMENTS = []

    PLAIN = "__plain__"

    @classmethod
    def isgroup(cls):
        return cls.__name__.startswith("Group")

    @classmethod
    def name(cls):
        return mini_buildd.misc.Snake(cls.__name__).from_camel()

    @classmethod
    def doc(cls):
        return inspect.getdoc(cls)

    @classmethod
    def arg_identities(cls):
        return [a.identity for a in cls.ARGUMENTS]

    @classmethod
    def api_args(cls, args):
        """
        Safely generate API args from args dict.

        - Use only known argument identities.
        - Use only non-None values.
        - Convert values to raw 'str' representation.
        """
        return {a.identity: a._v2r(args[a.identity]) for a in cls.ARGUMENTS if args[a.identity] is not None}   # pylint: disable=protected-access

    @classmethod
    def auth_err(cls, user):
        """
        Check if django user is authorized to call command.

        Empty string means user is authorized.
        """
        def chk_login():
            return user.is_authenticated and user.is_active

        if user is None:
            return f"API: '{cls.name()}': Internal Error: No user information available"
        if (cls.AUTH == cls.LOGIN) and not chk_login():
            return f"API: '{cls.name()}': Please login to run this command"
        if (cls.AUTH == cls.STAFF) and not (chk_login() and user.is_staff):
            return f"API: '{cls.name()}': Please login as 'staff' user to run this command"
        if (cls.AUTH == cls.ADMIN) and not (chk_login() and user.is_superuser):
            return f"API: '{cls.name()}': Please login as superuser ('admin') to run this command"
        return ""  # Auth OK

    def __init__(self, given_args=None, request=None):
        self.args = {}
        for arg in self.ARGUMENTS:
            self.args[arg.identity] = copy.copy(arg)

        self.request = request
        self.result = {}  # json result

        if given_args:
            # LOG.debug(f"Given args: {given_args}")
            for key in given_args:
                if key in self.args:
                    if isinstance(given_args[key], Argument):
                        self.args[key] = given_args[key]
                    elif isinstance(given_args, django.http.request.QueryDict):
                        # django request.GET args. Only consider non-empty values.
                        self.args[key].set(",".join([v for v in given_args.getlist(key) if v]))
                    else:
                        # dict arg (like from get_default_args)
                        self.args[key].set(given_args[key])
                    # LOG.debug(f"Arg updated: {self.args[key].identity}={self.args[key].value}")

        for arg in self.args.values():
            arg.update()

        self.update()

    @classmethod
    def json_pretty(cls, result):
        return json.dumps(result, indent=2)

    @classmethod
    def get_plain(cls, result):
        return result.get(cls.PLAIN, cls.json_pretty(result))

    def plain(self):
        return self.get_plain(self.result)

    def update(self):
        pass

    def run(self):
        # Sanity checks
        for argument in self.args.values():
            if argument.raw_value is None:
                raise mini_buildd.HTTPBadRequest(f"Missing required argument '{argument.identity}'")

        # Run
        self._run()

    def _run(self):
        raise Exception(f"No _run() function defined for: {self.name()}")

    def has_flag(self, flag):
        return self.args.get(flag, "False") == "True"


def _pimpdoc(result_doc=None):
    def pimp(cls):
        indent = "    "  # Be sure to keep the same indent like a top level class doc (4 chars) for all lines, so inspect.getdoc() will properly unindent.
        if result_doc:
            cls.__doc__ += "\n" + "\n".join([f"{indent}{line}" for line in result_doc]) + "\n"
        if not cls.isgroup():
            cls.__doc__ += f"\n{indent}Authorization: {cls.AUTH_STRINGS[cls.AUTH]}\n"
        return cls
    return pimp


class GroupSetup(Command):
    """
    Setup.
    """

    AUTH = Command.ADMIN
    CONFIRM = True


@_pimpdoc()
class SetupDaemon(GroupSetup):
    """Daemon setup."""

    def _run(self):
        http_endpoint = mini_buildd.http_endpoint()
        if http_endpoint.is_ssl():
            port = http_endpoint.options.get("port")
            mini_buildd.Daemon().model.ftpd_bind = http_endpoint.description.replace(f"port={port}", f"port={int(port)+1}")
            mini_buildd.Daemon().model.save(update_fields=["ftpd_bind"], mbd_sync=True)
        mini_buildd.Daemon().meta("daemon.Daemon", "pca_all", request=self.request)
        mini_buildd.Daemon().keyrings.set_needs_update()  # TODO: For now, we need this.


@_pimpdoc()
class SetupSources(GroupSetup):
    """Sources setup."""

    ARGUMENTS = [
        MultiSelectArgument(["--vendors", "-V"], default=["debian"], choices=["debian", "ubuntu"], doc="vendors to setup for"),
        MultiSelectArgument(["--archives", "-A"], default=[], doc="add arbitrary archives"),
        BoolArgument(["--with-vendor-archives"], default=True, doc="add original vendor archives"),
        BoolArgument(["--with-local-archives"], default=True, doc="add archives from your local sources.list and try to detect local apt-cacher-ng"),
    ]

    def _run(self):
        for url in self.args["archives"].value:
            from mini_buildd.models.source import Archive
            Archive.mbd_get_or_create(self.request, url=url)

        if self.args["with_local_archives"].value:
            mini_buildd.Daemon().meta("source.Archive", "add_local", request=self.request)
        for v in self.args["vendors"].value:
            if self.args["with_vendor_archives"].value:
                mini_buildd.Daemon().meta("source.Archive", f"add_{v}", request=self.request)
            mini_buildd.Daemon().meta("source.Source", f"add_{v}", request=self.request)
        mini_buildd.Daemon().meta("source.PrioritySource", "add_extras", request=self.request)
        mini_buildd.Daemon().meta("source.Source", "pca_all", request=self.request)


@_pimpdoc()
class SetupRepositories(GroupSetup):
    """Repositories setup."""

    ARGUMENTS = [MultiSelectArgument(["--repositories", "-R"], default=["test"], choices=["test", "debdev"], doc="repositories to setup for")]

    def _run(self):
        mini_buildd.Daemon().meta("distribution.Layout", "create_defaults", request=self.request)
        mini_buildd.Daemon().meta("distribution.Distribution", "add_base_sources", request=self.request)
        for r in self.args["repositories"].value:
            mini_buildd.Daemon().meta("repository.Repository", f"add_{r}", request=self.request)
        mini_buildd.Daemon().meta("repository.Repository", "pca_all", request=self.request)


@_pimpdoc()
class SetupChroots(GroupSetup):
    """Chroots setup."""

    ARGUMENTS = [SelectArgument(["--chroot-backend", "-C"], default="Dir", choices=["Dir", "File", "LVM", "LoopLVM", "BtrfsSnapshot"], doc="chroot backend to use, or empty string to not create chroots.")]

    def _run(self):
        if self.args["chroot_backend"].value:
            cb_class = f"chroot.{self.args['chroot_backend'].value}Chroot"
            mini_buildd.Daemon().meta(cb_class, "add_base_sources", request=self.request)
            mini_buildd.Daemon().meta(cb_class, "pca_all", request=self.request)


@_pimpdoc()
class Setup(GroupSetup):
    """
    Bootstrap setup.

    >>> Setup({'chroot_backend': 'File'}).args['chroot_backend'].value
    'File'
    >>> Setup({'chroot_backend': SelectArgument(["--chroot-backend", "-C"], default="BtrfsSnapshot", choices=["Dir", "File", "LVM", "LoopLVM", "BtrfsSnapshot"])}).args['chroot_backend'].value
    'BtrfsSnapshot'
    """

    ARGUMENTS = SetupDaemon.ARGUMENTS + SetupSources.ARGUMENTS + SetupRepositories.ARGUMENTS + SetupChroots.ARGUMENTS

    def _run(self):
        mini_buildd.Daemon().mbd_stop()

        SetupDaemon(self.args, request=self.request).run()
        SetupSources(self.args, request=self.request).run()
        SetupRepositories(self.args, request=self.request).run()
        SetupChroots(self.args, request=self.request).run()

        mini_buildd.Daemon().mbd_start()


class GroupDaemon(Command):
    """
    Daemon.
    """

    _STATUS_RESULT_DOC = (
        "JSON Result (dict):",
        "  identity    : string: instance identity",
        "  version     : string: mini-buildd version",
        "  http        : string: HTTP URL",
        "  ftp         : string: FTP URL",
        "  running     : bool: Whether instance is accepting incoming",
        "  load        : float: machine load (0 =< load <= 1+)",
        "  chroots     : list: [<codename>:<arch>, ..]: Active chroots",
        "  repositories: dict: {identity: [codename, ..]}: Active repositories",
        "  remotes     : list: Active or auto-reactivate remotes",
        "  packaging   : list: Active (source) packages being processed",
        "  building    : list: Active (binary) package builds")

    PORT_RESULT_DOC = (
        "JSON Result (dict):",
        "  uploaded: list of triples of all items uploaded")

    @classmethod
    def _status_result(cls):
        chroots = {}
        for c in mini_buildd.Daemon().get_active_chroots():
            chroots.setdefault(c.source.codename, [])
            chroots[c.source.codename].append(c.architecture.name)

        repositories = {}
        for r in mini_buildd.Daemon().get_active_repositories():
            repositories[r.identity] = {}
            repositories[r.identity]["layout"] = {s.suite.name: {"uploadable": s.uploadable,
                                                                 "experimental": s.experimental,
                                                                 "migrates_to": s.migrates_to.suite.name if s.migrates_to else None} for s in r.layout.suiteoption_set.all()}
            repositories[r.identity]["codenames"] = [d.base_source.codename for d in r.distributions.all()]

        return {
            "identity": mini_buildd.Daemon().model.identity,
            "version": mini_buildd.__version__,
            "http": mini_buildd.http_endpoint().geturl(),
            "ftp": mini_buildd.Daemon().model.mbd_get_ftp_endpoint().geturl(),
            "running": mini_buildd.Daemon().is_alive(),
            "load": mini_buildd.Daemon().builder.load(),
            "chroots": [c.mbd_key() for c in mini_buildd.Daemon().get_active_chroots()],
            "repositories": repositories,
            "remotes": [r.mbd_url() for r in mini_buildd.Daemon().get_active_or_auto_reactivate_remotes()],
            "packaging": [f"{p}" for p in list(mini_buildd.Daemon().packager.packaging.values())],
            "building": [f"{b}" for b in mini_buildd.Daemon().builder.alive()],
            "crontab": len(mini_buildd.Daemon().crontab.jobs),
            "events": len(mini_buildd.Daemon().events)}


@_pimpdoc(GroupDaemon._STATUS_RESULT_DOC)
class Status(GroupDaemon):
    """
    Show the status of the mini-buildd instance.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _run(self):
        self.result = self._status_result()
        # MsgLog(LOG, self.request).test_calls()

    @classmethod
    def has_chroot(cls, result, codename, arch):
        return mini_buildd.misc.chroot_key(codename, arch) in result["chroots"]


@_pimpdoc(GroupDaemon._STATUS_RESULT_DOC)
class Activate(GroupDaemon):
    """
    Activate the Daemon.
    """

    AUTH = Command.ADMIN

    def _run(self):
        from mini_buildd.models.daemon import Daemon as DaemonModel
        DaemonModel.Admin.mbd_activate(self.request, mini_buildd.Daemon().model)
        self.result = self._status_result()


@_pimpdoc(GroupDaemon._STATUS_RESULT_DOC)
class Deactivate(GroupDaemon):
    """
    Deactivate the Daemon.
    """

    AUTH = Command.ADMIN
    ARGUMENTS = []

    def _run(self):
        from mini_buildd.models.daemon import Daemon as DaemonModel
        DaemonModel.Admin.mbd_deactivate(self.request, mini_buildd.Daemon().model)
        self.result = self._status_result()


@_pimpdoc()
class TerminatePackage(GroupDaemon):
    """Terminate an ongoing package."""

    AUTH = Command.STAFF
    NEEDS_RUNNING_DAEMON = True
    CONFIRM = True
    ARGUMENTS = [TextArgument(["key"], doc="key (<distribution>_<package>_<version>[:<arch>]) of ongoing item.")]

    def _run(self):
        mini_buildd.Daemon().packager.terminate(self.args["key"].value)


@_pimpdoc()
class TerminateBuild(TerminatePackage):
    """Terminate an ongoing build."""

    def _run(self):
        mini_buildd.Daemon().builder.terminate(self.args["key"].value)


@_pimpdoc()
class GetKey(GroupDaemon):
    """
    Get GnuPG public key.

    JSON Result (dict):
      __plain__: string: ASCII-armored key.
    """

    def _run(self):
        self.result[self.PLAIN] = mini_buildd.Daemon().model.mbd_get_pub_key()


@_pimpdoc()
class GetDputConf(GroupDaemon):
    """
    Get recommended dput config snippet.

    Usually, this is for integration in your personal ~/.dput.cf.

    JSON Result (dict):
      __plain__: string: dput config snippet.
    """

    def _run(self):
        self.result[self.PLAIN] = mini_buildd.Daemon().model.mbd_get_dput_conf()


@_pimpdoc()
class GetUploaders(GroupDaemon):
    """
    Get all GPG keys allowed to upload to repositories.

    JSON Result (dict): {repository: {'allow_unauthenticated_uploads': True|False, 'uploaders': {long_key_id: {key info...}), ..}}}
    """

    AUTH = Command.ADMIN
    NEEDS_RUNNING_DAEMON = True
    ARGUMENTS = [RepositoryRegex(["--repository", "-R"])]

    def _run(self):
        mini_buildd.Daemon().keyrings.set_needs_update()
        for r in mini_buildd.Daemon().get_active_repositories().filter(identity__regex=rf"^{self.args['repository'].value}$"):
            self.result[r.identity] = {"allow_unauthenticated_uploads": r.allow_unauthenticated_uploads}
            self.result[r.identity]["uploaders"] = mini_buildd.Daemon().keyrings.get_uploaders()[r.identity].get_pub_keys_infos()


@_pimpdoc()
class LogCat(GroupDaemon):
    """
    Cat last n lines of the mini-buildd's log.

    JSON Result (dict):
      __plain__: string: last n loglines
    """

    AUTH = Command.STAFF
    ARGUMENTS = [
        NumberArgument(["--lines", "-n"], default=100, doc="cat (approx.) the last N lines")
    ]

    def _run(self):
        self.result[self.PLAIN] = mini_buildd.Daemon().log_cat(lines=self.args["lines"].value)


class GroupRepository(Command):
    """Repositories."""

    @classmethod
    def _upload_template_package(cls, template_package, dist):
        """Portext macro. Used for keyring_packages and test_packages."""
        with contextlib.closing(template_package) as package:
            dsc_url = "file://" + package.dsc
            info = f"Port for {dist}: {os.path.basename(dsc_url)}"
            try:
                return mini_buildd.Daemon().port_ext(dsc_url, dist)
            except BaseException as e:
                mini_buildd.log_exception(LOG, f"FAILED: {info}", e)


@_pimpdoc()
class GetSourcesList(GroupRepository):
    """
    Get sources.list (apt lines).

    Usually, this output is put to a file like '/etc/sources.list.d/mini-buildd-xyz.list'.

    JSON Result (dict):
      __plain__: string: apt lines
    """

    ARGUMENTS = [
        Codename(["codename"]),
        RepositoryRegex(["--repository", "-R"]),
        Suite(["--suite", "-S"]),
        MultiSelectArgument(["--types", "-T"], default=["deb"], choices=["deb", "deb-src"], doc="types of apt lines."),
        TextArgument(["--options", "-O"], default="", doc="apt line options (see 'man 5 source.list')"),
        NumberArgument(["--rollbacks", "-r"], default=0, doc="also list n rollback sources (0=no rollbacks)."),
        TextArgument(["--snapshot", "-P"], default="", doc="Select a repository snapshot."),
        TextArgument(["--mirror", "-M"], default="", doc="URL of a mirror to use (instead of mini-buildd's native URL)."),
        BoolArgument(["--with-extra", "-X"], default=False, doc="also list extra sources needed."),
        BoolArgument(["--with-comment", "-C"], default=False, doc="add comment line above apt line.")
    ]

    def _run(self):
        sources_list = mini_buildd.files.SourcesList()

        for r in mini_buildd.models.repository.Repository.objects.filter(identity__regex=fr"^{self.args['repository'].value}$"):
            for d in r.distributions.all().filter(base_source__codename__exact=self.args["codename"].value):
                if self.args["with_extra"].value:
                    for e in d.extra_sources.all():
                        sources_list.append(e.source.mbd_get_apt_line())

                for s in r.layout.suiteoption_set.filter(suite__name__regex=fr"^{self.args['suite'].value}$"):
                    sources_list.append(r.mbd_get_apt_line(d, s, snapshot=self.args["snapshot"].value))
                    if self.args["rollbacks"].value:
                        for rollback in range(0, self.args["rollbacks"].value):
                            sources_list.append(r.mbd_get_apt_line(d, s, snapshot=self.args["snapshot"].value, rollback=rollback))

        self.result[self.PLAIN] = sources_list.get(self.args["types"].value,
                                                   mirror=self.args["mirror"].value,
                                                   options=self.args["options"].value,
                                                   with_comment=self.args["with_comment"].value)


@_pimpdoc()
class GetSnapshots(GroupRepository):
    """
    Get list of repository snapshots for a distribution.

    JSON Result (dict):
      dict: <distribution>: [snasphot,...]: List of snapshots for the given distribution.
    """

    NEEDS_RUNNING_DAEMON = True
    ARGUMENTS = [
        Distribution(["distribution"]),
    ]

    def _run(self):
        repository, _distribution, _suite, _rollback = mini_buildd.models.repository.parse_distribution(self.args["distribution"].value)
        self.result = {self.args["distribution"].value: repository.mbd_reprepro.get_snapshots(self.args["distribution"].value)}


@_pimpdoc()
class GenSnapshot(GetSnapshots):
    """
    Generate a repository snapshot.

    JSON Result (dict): Empty for now.
    """

    AUTH = Command.ADMIN
    CONFIRM = True
    ARGUMENTS = GetSnapshots.ARGUMENTS + [TextArgument(["name"], doc="snapshot name.")]

    def _run(self):
        repository, _distribution, _suite, _rollback = mini_buildd.models.repository.parse_distribution(self.args["distribution"].value)
        repository.mbd_reprepro.gen_snapshot(self.args["distribution"].value, self.args["name"].value)


class DelSnapshot(GenSnapshot):
    """
    Delete a repository snapshot.

    JSON Result (dict): Empty for now.
    """

    def _run(self):
        repository, _distribution, _suite, _rollback = mini_buildd.models.repository.parse_distribution(self.args["distribution"].value)
        repository.mbd_reprepro.del_snapshot(self.args["distribution"].value, self.args["name"].value)


@_pimpdoc(GroupDaemon.PORT_RESULT_DOC)
class KeyringPackages(GroupRepository):
    """
    Build keyring packages for all active repositories.
    """

    AUTH = Command.ADMIN
    NEEDS_RUNNING_DAEMON = True
    CONFIRM = True
    ARGUMENTS = [
        KeyringDistributions(["--distributions", "-D"]),
        BoolArgument(["--with-migration", "-M"], default=True, doc="migrate packages."),
    ]

    def _run(self):
        self.result["uploaded"] = []
        events = mini_buildd.events.Attach(mini_buildd.Daemon().events)

        for d in self.args["distributions"].value:
            repository, distribution, suite, _rollback = mini_buildd.models.repository.parse_distribution(d)
            if not suite.build_keyring_package:
                raise mini_buildd.HTTPBadRequest(f"Keyring package to non-keyring suite requested (see 'build_keyring_package' flag): '{d}'")
            self.result["uploaded"].append(self._upload_template_package(mini_buildd.package.KeyringPackage(mini_buildd.Daemon().model), d))

        uploaded = set(self.result["uploaded"])
        while uploaded:
            event = events.get()
            for dist, source, version in uploaded.copy():
                if event.match(types=[mini_buildd.events.Type.INSTALLED, mini_buildd.events.Type.FAILED, mini_buildd.events.Type.REJECTED], distribution=dist, source=source, version=version):
                    LOG.info(f"Keyring package result: {event}")
                    uploaded.remove((dist, source, version))
                    if (event.type == mini_buildd.events.Type.INSTALLED) and self.args["with_migration"].value:
                        repository, distribution, suite, _rollback = mini_buildd.models.repository.parse_distribution(dist)
                        repository.mbd_package_migrate(source, distribution, suite, full=True, version=version, request=self.request)


@_pimpdoc(GroupDaemon.PORT_RESULT_DOC)
class TestPackages(GroupRepository):
    """
    Build internal test packages.
    """

    __TEMPLATES = ["mbd-test-archall", "mbd-test-cpp", "mbd-test-ftbfs"]

    AUTH = Command.ADMIN
    NEEDS_RUNNING_DAEMON = True
    CONFIRM = True
    ARGUMENTS = [
        MultiSelectArgument(["--sources", "-S"],
                            default=__TEMPLATES,
                            choices=__TEMPLATES,
                            doc="test source packages to use"),
        TestDistributions(["--distributions", "-D"]),
        AutoPortDistributions(["--auto-ports", "-A"]),
        BoolArgument(["--with-check", "-c"], default=False, doc="Check for correct packager results."),
    ]

    def _run(self):
        self.result["uploaded"] = []
        events = mini_buildd.events.Attach(mini_buildd.Daemon().events)

        for p in self.args["sources"].value:
            for d in self.args["distributions"].value:
                self.result["uploaded"].append(self._upload_template_package(mini_buildd.package.TestPackage(p, auto_ports=self.args["auto_ports"].value), d))

        if self.args["with_check"].value:
            uploaded = set(self.result["uploaded"])
            while uploaded:
                event = events.get()
                for dist, source, version in uploaded.copy():
                    if event.match(types=[mini_buildd.events.Type.INSTALLED, mini_buildd.events.Type.FAILED, mini_buildd.events.Type.REJECTED], distribution=dist, source=source, version=version):
                        uploaded.remove((dist, source, version))
                        if (event.type == mini_buildd.events.Type.FAILED and source in ["mbd-test-archall", "mbd-test-cpp"]) or \
                           (event.type == mini_buildd.events.Type.INSTALLED and source in ["mbd-test-ftbfs"]):
                            raise mini_buildd.HTTPBadRequest(f"Test package failed: {event}")
                        LOG.info(f"Test package result: {event}")


class GroupPackage(Command):
    """Packages."""


@_pimpdoc()
class List(GroupPackage):
    """
    List packages of a distribution, matching a shell-like glob pattern; matches both source and binary package names.

    JSON Result (dict): {repository: [{package_info}...]}
    """

    AUTH = Command.LOGIN
    ARGUMENTS = [
        Distribution(["distribution"]),
        PatternSource(["pattern"]),
        NumberArgument(["--max", "-M"], default=50, doc="max list size"),
    ]

    def _run(self):
        repository, _distribution, _suite, _rollback = mini_buildd.models.repository.parse_distribution(self.args["distribution"].value)
        self.result[repository.identity] = repository.mbd_reprepro.list(self.args["pattern"].value, self.args["distribution"].value, list_max=self.args["max"].value)


@_pimpdoc()
class Show(GroupPackage):
    """
    Show a source package.

    JSON Result (dict): {repository: [(codename, {package_info}),...]}
    """

    ARGUMENTS = [Source(["source"])]

    def _run(self):
        # Save all results of all repos in a top-level dict (don't add repos with empty results).
        for r in mini_buildd.Daemon().get_active_repositories():
            r_result = r.mbd_package_show(self.args["source"].value)
            if r_result:
                self.result[r.identity] = r_result


@_pimpdoc()
class Find(GroupPackage):
    """
    Find a source package.

    JSON Result (dict): {repository: [{package_info},...]}
    """

    ARGUMENTS = [
        Source(["source"]),
        FindDistributions(["--distributions", "-D"]),
        SourceVersion(["--version", "-V"]),
        SourceVersion(["--minimal-version", "-M"]),
    ]

    def _run(self):
        found = False
        for r in mini_buildd.Daemon().get_active_repositories():
            show = r.mbd_reprepro.show(self.args["source"].value)
            self.result[r.identity] = [s for s in show if
                                       (not self.args["distributions"].value or s["distribution"] in self.args["distributions"].value) and
                                       (not self.args["version"].value or s["version"] == self.args["version"].value) and
                                       (not self.args["minimal_version"].value or debian.debian_support.Version(s["version"]) >= debian.debian_support.Version(self.args["minimal_version"].value))]
            found = found or self.result[r.identity]

        if not found:
            raise mini_buildd.HTTPBadRequest(f"{self.args['source'].value} (version={self.args['version'].value or 'any'}) not found in {self.args['distributions'].value or 'any distribution'}")


@_pimpdoc()
class Migrate(GroupPackage):
    """
    Migrate a source package (along with all binary packages).

    If run for a rollback distribution, this will perform a rollback restore.

    JSON Result (dict):
      __plain__: string: textual migration log.
    """

    AUTH = Command.STAFF
    CONFIRM = True
    ARGUMENTS = [
        Source(["source"]),
        Distribution(["distribution"]),
        BoolArgument(["--full", "-F"], default=False, doc="migrate all 'migrates_to' suites up (f.e. unstable->testing->stable)."),
        SourceVersion(["--version", "-V"])
    ]

    def _run(self):
        repository, distribution, suite, rollback = mini_buildd.models.repository.parse_distribution(self.args["distribution"].value)
        self.result[self.PLAIN] = repository.mbd_package_migrate(self.args["source"].value,
                                                                 distribution,
                                                                 suite,
                                                                 full=self.args["full"].value,
                                                                 rollback=rollback,
                                                                 version=self.args["version"].false2none(),
                                                                 request=self.request)


@_pimpdoc()
class Remove(GroupPackage):
    """
    Remove a source package (along with all binary packages).

    JSON Result (dict):
      __plain__: string: textual removal log.
    """

    AUTH = Command.ADMIN
    CONFIRM = True
    ARGUMENTS = [
        Source(["source"]),
        Distribution(["distribution"]),
        BoolArgument(["--with-rollback"], default=True, doc="copy to rollback distribution"),
        SourceVersion(["--version", "-V"])
    ]

    def _run(self):
        repository, distribution, suite, rollback = mini_buildd.models.repository.parse_distribution(self.args["distribution"].value)
        self.result[self.PLAIN] = repository.mbd_package_remove(self.args["source"].value,
                                                                distribution,
                                                                suite,
                                                                rollback=rollback,
                                                                version=self.args["version"].false2none(),
                                                                request=self.request,
                                                                with_rollback=self.args["with_rollback"].value)


@_pimpdoc(GroupDaemon.PORT_RESULT_DOC)
class Port(GroupPackage):
    """
    Port an internal package.

    An internal 'port' is a no-changes (i.e., only the changelog
    will be adapted) rebuild of the given locally-installed
    package.

    When from-distribution equals to_distribution, a rebuild will be done.

    JSON Result (dict):
      requested: list: Info string about each port that has been triggered
    """

    AUTH = Command.STAFF
    NEEDS_RUNNING_DAEMON = True
    CONFIRM = True
    ARGUMENTS = [
        Source(["source"]),
        Distribution(["from_distribution"]),
        UploadableDistributions(["to_distributions"]),
        SourceVersion(["--version", "-V"]),
        UploadOptions(["--options", "-O"])]

    def update(self):
        if self.args["from_distribution"].value:
            repository, _distribution, suite, _rollback_no = mini_buildd.models.repository.parse_distribution(self.args["from_distribution"].value)
            self.args["to_distributions"].choices = repository.mbd_distribution_strings(uploadable=True, experimental=suite.experimental)

    def _run(self):
        self.result["uploaded"] = []
        for to_distribution in self.args["to_distributions"].value:
            self.result["uploaded"].append(mini_buildd.Daemon().port(self.args["source"].value,
                                                                     self.args["from_distribution"].value,
                                                                     to_distribution,
                                                                     version=self.args["version"].false2none(),
                                                                     options=self.args["options"].value.split("|")))


@_pimpdoc(GroupDaemon.PORT_RESULT_DOC)
class PortExt(GroupPackage):
    """
    Port an external package.

    An external 'port' is a no-changes (i.e., only the changelog
    will be adapted) rebuild of any given source package.

    JSON Result (dict):
      requested: list: Info string about each port that has been triggered.
    """

    AUTH = Command.STAFF
    NEEDS_RUNNING_DAEMON = True
    CONFIRM = True
    ARGUMENTS = [
        URLArgument(["dsc"], doc="URL of any Debian source package (dsc) to port"),
        UploadableDistributions(["distributions"]),
        UploadOptions(["--options", "-O"]),
    ]

    def _run(self):
        self.result["uploaded"] = []
        for d in self.args["distributions"].value:
            self.result["uploaded"].append(mini_buildd.Daemon().port_ext(self.args["dsc"].value, d, options=self.args["options"].value.split("|")))


@_pimpdoc()
class Retry(GroupPackage):
    """
    Retry a previously failed package.

    Use 'repository' in case of multiple matches only."

    JSON Result (dict):
      uploaded: string: Info string about each retry that has been triggered.
    """

    AUTH = Command.STAFF
    NEEDS_RUNNING_DAEMON = True
    CONFIRM = True
    ARGUMENTS = [
        Distribution(["distribution"]),
        Source(["source"]),
        SourceVersion(["version"]),
    ]

    def update(self):
        if self.args["source"].value:
            self.args["version"].choices = mini_buildd.Daemon().events.get_last_versions(self.args["source"].value)

    def _run(self):
        log = mini_buildd.changes.Log(self.args["distribution"].value, self.args["source"].value, self.args["version"].value)
        if log.installed():
            raise mini_buildd.HTTPBadRequest("Retry only works uninstalled source packages (can't retry)")
        mini_buildd.Daemon().incoming_queue.put(log.upload.file_path)
        self.result["uploaded"] = str(log.upload)


class GroupUser(Command):
    """Users."""


@_pimpdoc()
class SetUserKey(GroupUser):
    """
    Set a user's GnuPG public key.
    """

    AUTH = Command.LOGIN
    CONFIRM = True
    ARGUMENTS = [
        TextareaArgument(["key"], doc="GnuPG public key; multiline inputs will be handled as ascii armored full key, one-liners as key ids")
    ]

    def _run(self):
        uploader = self.request.user.uploader
        uploader.Admin.mbd_remove(self.request, uploader)
        key = self.args["key"].value

        if "\n" in key:
            MsgLog(LOG, self.request).info("Using given key argument as full ascii-armored GPG key")
            uploader.key_id = ""
            uploader.key = key
        else:
            MsgLog(LOG, self.request).info("Using given key argument as key ID")
            uploader.key_id = key
            uploader.key = ""

        uploader.Admin.mbd_prepare(self.request, uploader)
        uploader.Admin.mbd_check(self.request, uploader)
        MsgLog(LOG, self.request).info(f"Uploader profile changed: {uploader}")
        MsgLog(LOG, self.request).warning("Your uploader profile must be (re-)activated by the mini-buildd staff before you can actually use it.")


@_pimpdoc()
class Subscription(GroupUser):
    """
    Manage subscriptions to package notifications.

    A package subscription is a tuple 'PACKAGE:DISTRIBUTION',
    where both PACKAGE or DISTRIBUTION may be empty to denote
    all resp. items.

    JSON Result (dict):
      list: list: (action=list only) Human readable list of subscriptions.
      add: list: (action=add only) Human readable list of added subscriptions.
      remove: list: (action=lremoved only) Human readable list of removed subscriptions.
    """

    AUTH = Command.LOGIN
    ARGUMENTS = [
        SelectArgument(["action"], doc="action to run", choices=["list", "add", "remove"]),
        TextArgument(["subscription"], doc="subscription pattern")
    ]

    def update(self):
        if self.args["subscription"].value:
            self.args["subscription"].choices = [self.args["subscription"].value]
            package, _sep, _distribution = self.args["subscription"].value.partition(":")
            for r in mini_buildd.Daemon().get_active_repositories():
                self.args["subscription"].choices += [f"{package}:{d}" for d in r.mbd_distribution_strings()]

    def _run(self):
        package, _sep, distribution = self.args["subscription"].value.partition(":")

        def _filter():
            for s in mini_buildd.Daemon().get_subscription_objects().filter(subscriber=self.request.user):
                if package in ("", s.package) and distribution in ("", s.distribution):
                    yield s

        def _delete(subscription):
            result = f"{subscription}"
            subscription.delete()
            return result

        if self.args["action"].value == "list":
            self.result["list"] = [f"{subscription}." for subscription in _filter()]

        elif self.args["action"].value == "add":
            subscription, created = mini_buildd.Daemon().get_subscription_objects().get_or_create(subscriber=self.request.user,
                                                                                                  package=package,
                                                                                                  distribution=distribution)
            self.result["add"] = f"{'Added' if created else 'Exists'}: {subscription}."

        elif self.args["action"].value == "remove":
            self.result["remove"] = ["Removed: {s}.".format(s=_delete(subscription)) for subscription in _filter()]

        else:
            raise mini_buildd.HTTPBadRequest(f"Unknown action '{self.args['action'].value}': Use one of 'list', 'add' or 'remove'.")


class Commands(collections.OrderedDict):
    """Automatically collect all commands defined in this module, and make them accessible."""

    def __init__(self):
        super().__init__({c.name(): c for c in sys.modules[__name__].__dict__.values() if inspect.isclass(c) and issubclass(c, Command) and c != Command})

    def menu(self):
        """Menu-like structure of commands (helps creating the API menu in HTML)."""
        menu = []
        commands = None
        for _name, cmd in self.items():
            if cmd.isgroup():
                commands = []
                menu.append({"group": cmd(), "commands": commands})
            else:
                commands.append(cmd())
        return menu


COMMANDS = Commands()
