import os
import shutil
import re
import subprocess
import queue
import logging

import mini_buildd.config
import mini_buildd.misc
import mini_buildd.files
import mini_buildd.sbuild
import mini_buildd.threads
import mini_buildd.net
import mini_buildd.call
import mini_buildd.changes


LOG = logging.getLogger(__name__)


class Build(mini_buildd.threads.DeferredThread):
    def __init__(self, limiter, event):
        self.buildrequest = mini_buildd.changes.Buildrequest(event)
        super().__init__(limiter=limiter, name=f"Building {self.buildrequest}")

        self._chroot = f"mini-buildd-{self.buildrequest.distribution.codename}-{self.buildrequest['architecture']}"

        self.buildresult = self.buildrequest.gen_buildresult()

        # Always generate the now out-of-chroot 'libdir' (needed for ccache, ...).
        self._libdir = mini_buildd.config.HOME["chroots_lib"].path.get(str(self.buildrequest.distribution.codename), self.buildrequest["architecture"])
        os.makedirs(self._libdir, exist_ok=True)

        self.sbuildrc_path = os.path.join(self.buildrequest.get_spool_dir(), ".sbuildrc")

        self.config = mini_buildd.files.PerlModule(".sbuildrc", placeholders={"LIBDIR": self._libdir})
        self.setup = mini_buildd.files.Dir(os.path.join(self.buildrequest.get_spool_dir(), ".setup.d"))
        self.sbuild_cmd = []

        self.sbuild_call = None
        self.is_uploaded = None

    def __str__(self):
        return self.buildrequest.bkey

    def terminate(self):
        if self.is_alive() and self.sbuild_call:
            LOG.info(f"Terminating sbuild process for {self.buildrequest}...")
            self.sbuild_call.terminate()
            return True
        return False

    def built(self):
        return os.path.exists(self.buildresult.file_path)

    def _buildlog_to_buildresult(self, buildlog):
        """
        Parse sbuild's summary status from build log.

        .. note:: This will iterate all lines of the build log,
        and parse out the selection of sbuild's summary status
        we need.  In case the build log above does write the
        same output like 'Status: xyz', sbuild's correct status
        at the bottom will override this later.  Best thing,
        though, would be if sbuild would eventually provide a
        better way to get these values.
        """
        regex = re.compile("^(Status|Lintian|Piuparts|Autopkgtest): [^ ]+.*$")
        with mini_buildd.open_utf8(buildlog, errors="replace") as f:
            for line in f:
                if regex.match(line):
                    LOG.debug(f"Build log line detected as build status: {line.strip()}")
                    s = line.split(":")
                    self.buildresult.cset("Sbuild-" + s[0], s[1].strip())

    def generate_config(self):
        self.config.add("$path = '/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin:/usr/games';",
                        description="Always set $path, so config blocks may add to it (this is sbuild's default path)")
        mini_buildd.sbuild.CONFIG_BLOCKS.configure(self.config, self.buildrequest.cget, top=True)
        self.config.add("$apt_update = 0;",
                        description="Disable sbuild's internal apt update (we do a custom update later via chroot-setup commands)")
        self.config.add(f"$apt_allow_unauthenticated = {self.buildrequest.cget('Apt-Allow-Unauthenticated')};",
                        description="Allow unauthenticated apt toggle (from buildrequest)")
        mini_buildd.sbuild.CONFIG_BLOCKS.configure(self.config, self.buildrequest.cget, top=False)
        self.config.add_file(os.path.join(self.buildrequest.get_spool_dir(), "sbuildrc_snippet"),
                             description="Custom sbuild config (from Distribution)")
        self.config.save_as(self.sbuildrc_path)

    def generate_setup(self):
        sources_list_file = f"{self.buildrequest.get_spool_dir()}/apt_sources.list"

        self.setup.add(mini_buildd.files.ShScript("auto-show-sbuildrc",
                                                  description="Print used '.sbuildrc' (FYI)",
                                                  snippet=f"cat '{self.sbuildrc_path}'"))
        if self.buildrequest.cget('Apt-Allow-Unauthenticated', "0") == "1":
            self.setup.add(mini_buildd.files.ShScript("auto-apt-allow-unauthenticated",
                                                      description="Allow unauthenticated APT (as per Distribution.apt_allow_unauthenticated setting)",
                                                      snippet="printf 'APT::Get::AllowUnauthenticated \"true\";\\n' >'/etc/apt/apt.conf.d/10allow-unauthenticated'"))

        mini_buildd.sbuild.SETUP_BLOCKS.configure(self.setup, self.buildrequest.cget, top=True)
        if mini_buildd.files.SourcesList.has_https(sources_list_file):
            self.setup.add(mini_buildd.files.ShScript("auto-apt-https",
                                                      description="Enable APT https support",
                                                      snippet=f"""\
{mini_buildd.sbuild.APT_GET} install ca-certificates apt-transport-https
cp -v {self.buildrequest.get_spool_dir()}/ssl_cert /usr/local/share/ca-certificates/mini-buildd-repo.crt
/usr/sbin/update-ca-certificates
"""))

        self.setup.add(mini_buildd.files.ShScript("auto-apt-keys",
                                                  description="Setup APT keys from buildrequest",
                                                  snippet=f"""\
__install_gnupg()
{{
  # Workaround: Very old dists still may need to use apt-key (squeeze or older).
  # Workaround: Old dists may need to convert .asc to .gpg (jessie).
  # Workaround: Just using [trusted=yes] (>=wheezy) may work to obsolete all that, but seems unwise.
  {mini_buildd.sbuild.APT_GET} update
  {mini_buildd.sbuild.APT_GET} install gnupg
}}

APT_TRUSTED_GPGD="/etc/apt/trusted.gpg.d"
if [ -d "${{APT_TRUSTED_GPGD}}" ]; then
  cp -v "{self.buildrequest.get_spool_dir()}/apt_keys" "${{APT_TRUSTED_GPGD}}/mini-buildd-buildrequest.asc"
  case ${{SCHROOT_CHROOT_NAME}} in
    *-etch-*|*-lenny-*|*-squeeze-*|*-wheezy-*|*-jessie-*)
        echo "W: ${{SCHROOT_CHROOT_NAME}}: [apt <1.4/stretch] Only supports key.gpg (not key.asc) in ${{APT_TRUSTED_GPGD}}." >&2
        __install_gnupg
        gpg --dearmor "${{APT_TRUSTED_GPGD}}/mini-buildd-buildrequest.asc"
      ;;
  esac
else
  echo "W: ${{SCHROOT_CHROOT_NAME}}: Only supports deprecated apt-key." >&2
  __install_gnupg
  apt-key add "{self.buildrequest.get_spool_dir()}/apt_keys"
fi
"""))

        self.setup.add(mini_buildd.files.ShScript("auto-apt-update",
                                                  description="Setup APT sources and preferences, and update",
                                                  snippet=f"""\
cp -v "{sources_list_file}" /etc/apt/sources.list
cat /etc/apt/sources.list

cp -v "{self.buildrequest.get_spool_dir()}/apt_preferences" /etc/apt/preferences
cat /etc/apt/preferences

{mini_buildd.sbuild.APT_GET} update
apt-cache policy
"""))

        mini_buildd.sbuild.SETUP_BLOCKS.configure(self.setup, self.buildrequest.cget, top=False)
        custom_buildrequest = os.path.join(self.buildrequest.get_spool_dir(), "chroot_setup_script")
        if os.path.getsize(custom_buildrequest) > 0:
            self.setup.add(mini_buildd.files.AutoScript(custom_buildrequest))
        self.setup.save()

    def generate_sbuild_cmd(self):
        self.sbuild_cmd = ["sbuild",
                           "--dist", self.buildrequest["distribution"],
                           "--arch", self.buildrequest["architecture"],
                           "--chroot", self._chroot]

        for name in self.setup.keys():
            script = os.path.join(self.setup.path, name)
            self.sbuild_cmd += ["--chroot-setup-command", script]

        self.sbuild_cmd += ["--build-dep-resolver", self.buildrequest.cget("Build-Dep-Resolver"),
                            "--keyid", mini_buildd.Daemon().model.mbd_gnupg().get_first_sec_key(),
                            "--nolog",
                            "--log-external-command-output",
                            "--log-external-command-error"]

        if self.buildrequest.cget("Arch-All"):
            self.sbuild_cmd += ["--arch-all"]
        else:
            self.sbuild_cmd += ["--no-arch-all"]  # Be sure to set explicitly: sbuild >= 0.77 does not seem to use this as default any more?

        lintian_check = self.buildrequest.check_mode("lintian")
        if lintian_check.mode.value > 0:
            self.sbuild_cmd += ["--run-lintian"]
            self.sbuild_cmd += ["--lintian-opts", self.buildrequest.distribution.codename.lintian_options(lintian_check)]
            self.sbuild_cmd += ["--lintian-opts", self.buildrequest.check_extra_options("lintian")]
        else:
            self.sbuild_cmd += ["--no-run-lintian"]

        piuparts_check = self.buildrequest.check_mode("piuparts")
        if piuparts_check.mode.value > 0:
            self.sbuild_cmd += ["--run-piuparts"]
            self.sbuild_cmd += ["--piuparts-opts", f"--schroot {self._chroot} --dpkg-force-confdef --warn-on-debsums-errors --warn-on-leftovers-after-purge --warn-on-others --keep-sources-list"]
            self.sbuild_cmd += ["--piuparts-opts", self.buildrequest.check_extra_options("piuparts")]
        else:
            self.sbuild_cmd += ["--no-run-piuparts"]

        autopkgtest_check = self.buildrequest.check_mode("autopkgtest")
        if autopkgtest_check.mode.value > 0:
            self.sbuild_cmd += ["--run-autopkgtest"]
            self.sbuild_cmd += ["--autopkgtest-opts", f"-- schroot {self._chroot}"]
            self.sbuild_cmd += ["--autopkgtest-root-args", ""]
        else:
            self.sbuild_cmd += ["--no-run-autopkgtest"]

        if "sbuild" in mini_buildd.config.DEBUG:
            self.sbuild_cmd += ["--debug"]

        self.sbuild_cmd += [self.buildrequest.dsc_name]

    def build(self):
        mini_buildd.Daemon().keyrings.get_remotes().verify(self.buildrequest.file_path)
        self.buildrequest.untar(path=self.buildrequest.get_spool_dir())
        self.generate_config()
        self.generate_setup()
        self.generate_sbuild_cmd()

        # Actually run sbuild
        buildlog = os.path.join(self.buildrequest.get_spool_dir(), self.buildrequest.buildlog_name)
        live_buildlog = mini_buildd.config.HOME["spool"].path.get(self.buildrequest.live_buildlog_name)
        with open(buildlog, "w+") as buildlog_file:
            LOG.info(f"Adding live buildlog: {live_buildlog}")
            # The spool id/hash might the very same (retry a failed build, for example) as a previous one. So we need to be sure to remove before linking.
            # Note that this currently should never really happen as python's tarball abstraction cannot produce reproducible tarballs (https://bugs.python.org/issue24465).
            if os.path.exists(live_buildlog):
                os.remove(live_buildlog)
            os.link(buildlog, live_buildlog)
            self.sbuild_call = subprocess.Popen(self.sbuild_cmd,
                                                cwd=self.buildrequest.get_spool_dir(),
                                                env=mini_buildd.call.taint_env({"HOME": self.buildrequest.get_spool_dir(),
                                                                                "GNUPGHOME": mini_buildd.config.HOME.path.get(".gnupg"),  # pylint: disable=no-member
                                                                                "DEB_BUILD_OPTIONS": self.buildrequest.cget("Deb-Build-Options", ""),
                                                                                "DEB_BUILD_PROFILES": self.buildrequest.cget("Deb-Build-Profiles", "")}),
                                                stdout=buildlog_file, stderr=subprocess.STDOUT)
            retval = self.sbuild_call.wait()

        # Update buildresult
        self.buildresult.cset("Runtime-Start", self.running.start.isoformat())
        self.buildresult.cset("Runtime-Delta", str(self.running.delta().total_seconds()))
        self.buildresult.cset("Sbuild-Retval", str(retval))
        self._buildlog_to_buildresult(buildlog)

        LOG.info(f"{self}: Sbuild finished: Sbuild-Retval={retval}, Status={self.buildresult.cget('Sbuild-Status')}")
        self.buildresult.add_file(buildlog)
        build_changes_file = os.path.join(self.buildrequest.get_spool_dir(),
                                          mini_buildd.changes.Changes.gen_changes_file_name(self.buildrequest["source"],
                                                                                            self.buildrequest["version"],
                                                                                            self.buildrequest["architecture"]))
        if os.path.exists(build_changes_file):
            build_changes = mini_buildd.changes.Buildrequest(build_changes_file)
            build_changes.tar(tar_path=self.buildresult.file_path + ".tar")
            self.buildresult.add_file(self.buildresult.file_path + ".tar")

        self.buildresult.save(mini_buildd.Daemon().model.mbd_gnupg())

    def clean(self):
        mini_buildd.misc.skip_if_keep_in_debug(shutil.rmtree, self.buildrequest.get_spool_dir())
        self.buildrequest.remove()

    def run_deferred(self):
        if self._shutdown is mini_buildd.config.SHUTDOWN:
            raise mini_buildd.HTTPShutdown

        # Build if needed (may be just an upload-pending build)
        if not self.built():
            self.build()

        # Upload
        self.buildresult.upload(mini_buildd.net.ClientEndpoint(self.buildrequest.cget("Upload-To"), protocol="ftp"))
        self.is_uploaded = mini_buildd.misc.datetime_stamp()
        mini_buildd.Daemon().events.logchanges(mini_buildd.events.Type.BUILT, self.buildresult, extra=self.buildresult.to_json())
        # Clean. Unless this is done, buildresult stays in incoming and may be retried.
        self.clean()


class Builder(mini_buildd.threads.EventThread):
    def __init__(self, max_parallel_builds):
        self.queue = queue.Queue()
        self.limiter = queue.Queue(maxsize=max_parallel_builds)
        self.builds = {}
        super().__init__(events=self.queue, name="builder")

    def __str__(self):
        return f"Building {len(self.alive())}/{self.limiter.maxsize} (load {self.load()}, {len(self.builds)} threads)"

    def alive(self):
        return [build for build in self.builds.values() if build.is_alive()]

    def clear(self):
        with self.lock:
            for key in [key for key, build in self.builds.items() if not build.is_alive()]:  # Iterate on copy, del on actual dict
                self.builds[key].join()
                del self.builds[key]

    def join(self, timeout=None):
        for build in self.builds.values():
            build.terminate()
            build.join()
        super().join()

    def load(self):
        return round(float(len(self.alive()) / self.limiter.maxsize), 2)

    def run_event(self, event):
        self.clear()
        build = Build(self.limiter, event)
        self.builds[build.buildrequest.bkey] = build
        build.start()

    def terminate(self, key):
        self.clear()
        build = self.builds.get(key)
        if not (build and build.terminate()):
            raise mini_buildd.HTTPBadRequest(f"No such active build: {key}")
