import os
import logging
import json
import uuid
import re

import django.utils
import django
import mini_buildd
import mini_buildd.api


LOG = logging.getLogger(__name__)

register = django.template.Library()  # pylint: disable=invalid-name  # django needs this lower-case afaiu


@register.simple_tag
def mbd_get(dict_, key):
    """Get value from dict even if the key has a special name (like hyphens in a changes file field)."""
    return dict_.get(key)


@register.filter
def mbd_dirname(path):
    return os.path.dirname(path)


@register.filter
def mbd_basename(path):
    return os.path.basename(path)


@register.filter
def mbd_jsonpp(obj):
    """Get pretty print from json-serializable object."""
    return json.dumps(obj, indent=2)


@register.filter
def mbd_cssify(string):
    return string.replace("_", "-").replace(" ", "-").lower()


@register.simple_tag
def mbd_token():
    return f"mbd-token-{uuid.uuid4()}"


@register.simple_tag
def mbd_uri(name, variant, *args):
    return mini_buildd.config.URIS[name][variant].getabs(*args)


@register.simple_tag
def mbd_uri_home():
    return mbd_uri("home", "view")


@register.simple_tag(takes_context=True)
def mbd_next(context):
    request = context.get("request")
    return mbd_uri_home() if re.match(".*/snippets/.*", request.path) else request.get_full_path()


@register.inclusion_tag("includes/mbd_img.html")
def mbd_img(img, title, cls="mbd-icon", href=None, onclick=None, clipboard=None):
    return {"img": img, "title": title, "cls": cls, "href": href, "onclick": onclick, "clipboard": clipboard}


@register.inclusion_tag("includes/mbd_img.html")
def mbd_home():
    return mbd_img("mini_buildd", f"Home: {mini_buildd.Daemon().get_title()}", cls="mbd-logo", href=mbd_uri_home())


@register.simple_tag
def mbd_model_stats(model):
    ret = {}
    model_class = eval(f"mini_buildd.models.{model}")  # pylint: disable=eval-used
    if getattr(model_class, "mbd_is_prepared", None):
        # Status model
        ret["active"] = model_class.objects.filter(status__exact=model_class.STATUS_ACTIVE).count()
        ret["prepared"] = model_class.objects.filter(status__exact=model_class.STATUS_PREPARED).count()
        ret["removed"] = model_class.objects.filter(status__exact=model_class.STATUS_REMOVED).count()
    ret["total"] = model_class.objects.all().count()
    return ret


@register.inclusion_tag("includes/mbd_api_arg.html")
def mbd_api_arg(arg, token):
    return {"arg": arg, "token": token}


@register.inclusion_tag("includes/mbd_api.html", takes_context=True)
def mbd_api(context, cmd, name=None, title=None, full=True, action="ajax", **kwargs):
    def _kwargs(prefix):
        return {k[len(prefix):]: v for k, v in kwargs.items() if k.startswith(prefix)}

    if isinstance(cmd, str):
        api_cls = mini_buildd.api.COMMANDS.get(cmd)
        if api_cls is None:
            raise mini_buildd.HTTPInternal(f"Unknown command: {cmd}")
        api_cmd = api_cls(_kwargs("value_"))
    else:
        api_cmd = cmd
    auth_err = api_cmd.auth_err(context.get("user"))

    args = {"mandatory": [], "optional": [], "hidden": []}
    for arg in api_cmd.args.values():
        args["mandatory" if arg.value is None else "optional" if full else "hidden"].append(arg)

    return {"mbd": context.get("mbd"),
            "args": args,
            "api_cmd": api_cmd,
            "auth_err": auth_err,
            "name": name,
            "title": title,
            "action": action,
            "reload": mbd_next(context)}


@register.inclusion_tag("includes/mbd_sbuild_status.html")
def mbd_sbuild_status(bres):
    return {
        "bres": bres,
        "checks": {
            "lintian": bres.get("sbuild_lintian", "none"),
            "piuparts": bres.get("sbuild_piuparts", "none"),
            "autopkgtest": bres.get("sbuild_autopkgtest", "none"),
        }}


class Plain(dict):
    BUILDLOG_INDEX_REGEXES = [
        (re.compile(r"\| (.*) \|"), 1, "mbd-log-info"),  # sbuild section headers
        (re.compile(r"^MINI_BUILDD: .+"), 0, " mbd-log-info"),  # mini-buildd banners (see files.py)
        (re.compile(r".*make.*\*\*\*.*"), 0, " mbd-log-error"),  # make fatal error
        (re.compile(r"(^W(ARNING)?: .+)|(^W(ARN)?: .+)", re.IGNORECASE), 0, " mbd-log-warning"),  # warnings (lintian, others)
        (re.compile(r"(^E(RROR)?: .+)|(^E(RR)?: .+)", re.IGNORECASE), 0, " mbd-log-error"),  # errors (lintian, others)
    ]

    def buildlog_index(self, file_):
        log = ""
        for line in file_:
            for m in self.BUILDLOG_INDEX_REGEXES:
                match = m[0].match(line)
                if match:
                    token = mbd_token()
                    log += f"<span id='{token}'></span>"
                    title = django.utils.html.escape(match.group(m[1]))
                    self["toc"].append({"class": m[2], "token": token, "title": title})
                    break
            log += django.utils.html.escape(line)
        return django.utils.safestring.mark_safe(log)

    def cat(self, file_name, value_if_missing="File not found"):
        if os.path.exists(file_name):
            with open(file_name, "r", encoding=mini_buildd.config.CHAR_ENCODING, errors="ignore") as f:
                if file_name.endswith(".buildlog"):
                    return self.buildlog_index(f)
                return f.read()
        return value_if_missing

    def __init__(self, content, file_path, download_uri):
        super().__init__()
        self["file_path"] = file_path
        self["download_uri"] = download_uri
        self["toc"] = []
        self["content"] = self.cat(file_path) if file_path else content


@register.inclusion_tag("includes/mbd_plain.html")
def mbd_plain(content=None, file_path=None, download_uri=None):
    return Plain(content, file_path, download_uri)


@register.inclusion_tag("includes/mbd_snippet.html")
def mbd_snippet(name, items, show_ongoing=False):
    return {"name": name, "items": items, "show_ongoing": show_ongoing}


@register.simple_tag
def mbd_parse_distribution(repo, dist, suite):
    name = repo.mbd_get_distribution_string(dist, suite)
    return {"name": name,
            "uri": mini_buildd.config.HOME["repositories"].uris["static"].get(repo.identity, "dists", name + "/"),
            "mandatory_version_regex": repo.mbd_get_mandatory_version_regex(dist, suite),
            "snapshots": repo.mbd_reprepro.get_snapshots(name),
            "apt_build_sources_list": repo.mbd_get_apt_build_sources_list(dist, suite).get(with_comment=True),
            "apt_build_preferences": repo.mbd_get_apt_build_preferences(dist, suite)}
