import os
import stat
import glob
import fnmatch
import logging
import tarfile
import ftplib
import urllib.parse
import re
import contextlib

import debian.deb822

import mini_buildd.config
import mini_buildd.misc
import mini_buildd.distribution
import mini_buildd.net

LOG = logging.getLogger(__name__)


class LogPath():
    """
    Package log path: REPOID/PACKAGE/VERSION.
    """

    def __init__(self, repository, source, version, create=False):
        self.path = os.path.join(repository, source, version)
        self.fullpath = mini_buildd.config.HOME["events"].path.get(self.path)
        self.exists = os.path.isdir(self.fullpath)
        if not self.exists and create:
            os.makedirs(self.fullpath)


class Changes(debian.deb822.Changes):  # pylint: disable=too-many-ancestors
    FILE_APX = ""

    @classmethod
    def match(cls, file_path):
        return re.compile(f"^.+{cls.FILE_APX}_[^_]+.changes$").match(file_path)

    def _spool_hash_from_file(self):
        return None if not os.path.exists(self._file_path) else mini_buildd.misc.Hash(self._file_path).sha1()

    def __init__(self, file_path):
        self._file_path = file_path
        self._file_name = os.path.basename(file_path)
        self._new = not os.path.exists(file_path)
        # Instance might be produced from a temporary file, so we need to save the hash now.
        self._spool_hash = self._spool_hash_from_file()

        if self._new:
            super().__init__([])
        else:
            with mini_buildd.open_utf8(file_path) as cf:
                super().__init__(cf)

        # Be sure base dir is always available
        os.makedirs(os.path.dirname(file_path), exist_ok=True)

    @property
    def key(self):
        """Key to identify if a 'Package' (Upload, Buildrequest, Buildresult)."""
        return f"{self['source']}_{self['version']}@{self['distribution']}"

    @property
    def bkey(self):
        """Key to identify a 'Build' (Buildrequest, Buildresult)."""
        return f"{self.key}:{self['architecture']}"

    @property
    def distribution(self):
        return mini_buildd.distribution.Distribution(self["distribution"])

    def cget(self, field, default=None):
        return self.get(mini_buildd.misc.CField(field).fullname, default=default)

    def cset(self, field, value):
        self[mini_buildd.misc.CField(field).fullname] = value

    @classmethod
    def gen_changes_file_name(cls, package, version, arch, apx=""):
        """
        Gen any changes file name.

        Always strip epoch from version, and handle special
        mini-buildd types.

        >>> Changes.gen_changes_file_name("mypkg", "1.2.3-1", "mips")
        'mypkg_1.2.3-1_mips.changes'
        >>> Changes.gen_changes_file_name("mypkg", "7:1.2.3-1", "mips")
        'mypkg_1.2.3-1_mips.changes'
        >>> Changes.gen_changes_file_name("mypkg", "7:1.2.3-1", "mips", apx=Buildrequest.FILE_APX)
        'mypkg_1.2.3-1_mini-buildd-buildrequest_mips.changes'
        >>> Changes.gen_changes_file_name("mypkg", "7:1.2.3-1", "mips", apx=Buildresult.FILE_APX)
        'mypkg_1.2.3-1_mini-buildd-buildresult_mips.changes'
        """
        return f"{package}_{mini_buildd.misc.strip_epoch(version)}{apx}_{arch}.changes"

    def gen_file_name(self, arch, apx):
        return self.gen_changes_file_name(self["Source"], self["Version"], arch, apx)

    @classmethod
    def gen_dsc_file_name(cls, package, version):
        return f"{package}_{mini_buildd.misc.strip_epoch(version)}.dsc"

    @property
    def dsc_name(self):
        return self.gen_dsc_file_name(self["Source"], self["Version"])

    @property
    def dsc_file_name(self):
        return os.path.join(os.path.dirname(self._file_path), self.dsc_name)

    @property
    def file_name(self):
        return self._file_name

    @property
    def file_path(self):
        return self._file_path

    @property
    def buildlog_name(self):
        return f"{self['Source']}_{self['Version']}_{self['Architecture']}.buildlog"

    def get_log_path(self):
        return LogPath(self["Distribution"], self["Source"], self["Version"])

    def is_new(self):
        return self._new

    def get_spool_id(self):
        return f"{type(self).__name__}-{self._spool_hash}"

    def get_spool_dir(self):
        return mini_buildd.config.HOME["spool"].path.get(self.get_spool_id())

    def get_files(self, key=None):
        return [f[key] if key else f for f in self.get("Files", [])]

    def add_file(self, file_name):
        self.setdefault("Files", [])
        self["Files"].append({"md5sum": mini_buildd.misc.Hash(file_name).md5(),
                              "size": os.path.getsize(file_name),
                              "section": "mini-buildd",
                              "priority": "extra",
                              "name": os.path.basename(file_name)})

    def save(self, gnupg=None):
        """
        Write to file (optionally signed).

        >>> import tempfile
        >>> t = tempfile.NamedTemporaryFile()
        >>> c = Changes(t.name)
        >>> c["key"] = "ASCII value"
        >>> c.save(None)
        >>> c["key"] = "Ünicöde «value»"
        >>> c.save(None)
        """
        try:
            LOG.info(f"Saving changes: {self._file_path}")
            with open(self._file_path, "w+", encoding=mini_buildd.config.CHAR_ENCODING) as f:
                f.write(self.dump())

            LOG.info(f"Signing changes: {self._file_path}")
            if gnupg:
                gnupg.sign(self._file_path)
            self._spool_hash = self._spool_hash_from_file()
        except BaseException:
            # Existence of the file name is used as flag
            if os.path.exists(self._file_path):
                os.remove(self._file_path)
            raise

    def _upload(self, FTP, endpoint, force=False):  # pylint: disable=invalid-name  # noqa (pep8 N802)
        pre_upload_timestamp = mini_buildd.misc.datetime_stamp()

        upload = os.path.splitext(self._file_path)[0] + ".upload"
        if not force and os.path.exists(upload):
            with mini_buildd.open_utf8(upload) as uf:
                LOG.warning(f"FTP: '{upload}' exists (already uploaded to '{uf.read()}')")
        else:
            with FTP() as ftp:
                ftp.connect(endpoint.options.get("host"), int(endpoint.options.get("port")))
                if isinstance(ftp, ftplib.FTP_TLS):
                    ftp.auth()
                    ftp.prot_p()
                ftp.login()
                ftp.cwd("/incoming")
                for fd in self.get_files() + [{"name": self._file_name}]:
                    f = fd["name"]
                    LOG.debug(f"FTP: Uploading file: '{f}'")
                    with open(os.path.join(os.path.dirname(self._file_path), f), "rb") as fi:
                        ftp.storbinary(f"STOR {f}", fi)
                with mini_buildd.open_utf8(upload, "w") as fi:
                    fi.write(f"{endpoint.geturl()}")
                LOG.info(f"FTP: '{self._file_name}' uploaded to '{endpoint.geturl()}'.")

        return pre_upload_timestamp  # for events --since=

    def upload(self, endpoint, force=False):
        return self._upload(ftplib.FTP_TLS if endpoint.is_ssl() else ftplib.FTP, endpoint, force=force)

    def tar(self, tar_path, add_files=None, exclude_globs=None):
        exclude_globs = exclude_globs if exclude_globs else []

        def exclude(file_name):
            for e in exclude_globs:
                if fnmatch.fnmatch(file_name, e):
                    return True
            return False

        with contextlib.closing(tarfile.open(tar_path, "w")) as tar:
            def tar_add(file_name):
                if exclude(file_name):
                    LOG.info(f"Excluding \"{file_name}\" from tar archive \"{tar_path}\".")
                else:
                    tar.add(file_name, arcname=os.path.basename(file_name))

            tar_add(self._file_path)
            for f in self.get_files():
                tar_add(os.path.join(os.path.dirname(self._file_path), f["name"]))
            if add_files:
                for f in add_files:
                    tar_add(f)

    def untar(self, path):
        tar_file = self._file_path + ".tar"
        if os.path.exists(tar_file):
            with contextlib.closing(tarfile.open(tar_file, "r")) as tar:
                tar.extractall(path=path)
        else:
            LOG.info(f"No tar file (skipping): {tar_file}")

    def move_to_pkglog(self, installed, rejected=False):
        logdir = None if rejected else self.get_log_path().fullpath

        if logdir and not os.path.exists(logdir):
            os.makedirs(logdir)

        LOG.info(f"Moving changes to package log: '{self._file_path}'->'{logdir}'")
        for fd in [{"name": self._file_name}] + self.get_files():
            f = fd["name"]
            f_abs = os.path.join(os.path.dirname(self._file_path), f)
            # If not installed, just move all files to log dir.
            # If installed, only save buildlogs and changes.
            if logdir and (not installed or re.match(r"(.*\.buildlog$|.*changes$)", f)):
                LOG.info(f"Moving '{f}' to '{logdir}'")
                os.rename(f_abs, os.path.join(logdir, f))
            else:
                LOG.info(f"Removing '{f}'")
                mini_buildd.misc.skip_if_keep_in_debug(os.remove, f_abs)

    def remove(self):
        LOG.info(f"Removing changes: '{self._file_path}'")
        for fd in [{"name": self._file_name}] + self.get_files():
            f = os.path.join(os.path.dirname(self._file_path), fd["name"])
            LOG.debug(f"Removing: '{fd['name']}'")
            os.remove(f)

    def fields(self):
        return {mini_buildd.misc.Field(field): value for field, value in self.items()}

    def to_json(self):
        return {field.snake_name: value for field, value in self.fields().items() if field.is_cfield}


class Upload(Changes):  # pylint: disable=too-many-ancestors
    class Options():
        """
        Uploader options in changes.

        >>> f"{Upload('test-data/changes.options').options}"
        "auto-ports=['jessie-test-unstable', 'squeeze-test-snasphot'], autopkgtest-mode=errfail, internal-apt-priority=543, lintian-mode=errfail, piuparts-mode=errfail"

        >>> f"{Upload('test-data/changes.options_1.0').options}"
        "auto-ports=['jessie-test-unstable', 'squeeze-test-snasphot'], internal-apt-priority=543, lintian-mode=errfail, lintian-mode[i386]=disabled"

        >>> f"{Upload('test-data/changes.options_0.8').options}"
        "auto-ports=['jessie-test-unstable', 'squeeze-test-snasphot'], lintian-mode=ignore"
        """

        class Bool():
            _TRUE = ["true", "1"]
            _FALSE = ["false", "0"]
            _VALID = _TRUE + _FALSE

            def __init__(self, raw_value):
                if raw_value.lower() not in self._VALID:
                    raise Exception(f"Bool value must be one of {','.join(self._VALID)}")
                self.value = raw_value.lower() in self._TRUE

        class Int():
            def __init__(self, raw_value):
                self.value = int(raw_value)

        class Str():
            def __init__(self, raw_value):
                self.value = str(raw_value)

        class CSV():
            def __init__(self, raw_value):
                self.value = raw_value.split(",")

        KEYWORD = "MINI_BUILDD_OPTION"

        _OPTIONS_DEPRECATED = {
            "ignore-lintian": Bool,
            "run-lintian": Bool,
        }

        _OPTIONS = {**{
            "lintian-mode": Str,
            "piuparts-mode": Str,
            "autopkgtest-mode": Str,
            "internal-apt-priority": Int,
            "auto-ports": CSV,
            "deb-build-options": Str,
            "deb-build-profiles": Str,
        }, **_OPTIONS_DEPRECATED}

        @classmethod
        def usage(cls):
            return ("(lintian|piuparts|autopkgtest)-mode[<arch>]=<mode>\n"
                    f"  {mini_buildd.distribution.SbuildCheck.usage()}\n"
                    "internal-apt-priority[<arch>]=<prio>\n"
                    "  Apt prio for our repo packages when building (use 500 to always prefer).\n"
                    "auto-ports=<dist>,...\n"
                    "  List of distributions (comma-separated) to automatically run ports for after successful install.\n"
                    "deb-build-options[<arch>]=<option> <option>...\n"
                    "  Contents of DEB_BUILD_OPTIONS environment for building (see https://www.debian.org/doc/debian-policy/ch-source.html#debian-rules-and-deb-build-options).\n"
                    "deb-build-profiles[<arch>]=<profile> <profile>...\n"
                    "  Contents of DEB_BUILD_PROFILES environment for building (see https://wiki.debian.org/BuildProfileSpec).\n"
                    )

        @classmethod
        def _get_top_changes(cls, upload_changes):
            """
            Filter only the first block from the changes (changelog) entry.

            Upload changes may include multiple version blocks from
            the changelog (internal porting does it, for example),
            but we must only consider values from the top one.
            """
            result = ""
            header_found = False
            for line in upload_changes.get("Changes", "").splitlines(True):
                if re.match(r"^ [a-z0-9]+", line):
                    if header_found:
                        break
                    header_found = True
                result += line
            return result

        def _parse_ancient(self):
            """Compat parse support for ancient 'magic' options."""
            def warning(ancient, option):
                LOG.warning(f"Ancient user option \"{ancient}\" found. Please use option \"{option}\" instead (see user manual).")

            ancient_auto_backports = re.search(r"\*\s*MINI_BUILDD:\s*AUTO_BACKPORTS:\s*([^*.\[\]]+)", self._top_changes)
            if ancient_auto_backports:
                warning("AUTO_BACKPORTS", "auto-ports")
                self._set("auto-ports", ancient_auto_backports.group(1))

            ancient_backport_mode = re.search(r"\*\s*MINI_BUILDD:\s*BACKPORT_MODE", self._top_changes)
            if ancient_backport_mode:
                warning("BACKPORT_MODE", "lintian-mode")
                self._set("lintian-mode", "ignore")

        def __init__(self, upload_changes):
            self._top_changes = self._get_top_changes(upload_changes)
            self._options = {}
            matches = re.findall(fr"\*\s*{self.KEYWORD}:\s*([^*.]+)=([^*.]+)", self._top_changes)
            for m in matches:
                self._set(m[0], m[1])

            self._parse_ancient()

        def __str__(self):
            return ", ".join(f"{key}={value}" for key, value in sorted(self._options.items()))

        def _set(self, key, raw_value):
            key_part = key.partition("[")
            base_key = key_part[0]
            value = re.sub(r"\s+", "", raw_value)

            # Validity check for key
            if base_key not in list(self._OPTIONS.keys()):
                raise mini_buildd.HTTPBadRequest(f"Unknown upload option: {key}")

            # Duplicity check
            if key in list(self._options.keys()):
                raise mini_buildd.HTTPBadRequest(f"Duplicate upload option: {key}")

            # Value conversion check
            converted_value = None
            try:
                converted_value = self._OPTIONS[base_key](value)
            except BaseException:
                raise mini_buildd.HTTPBadRequest(f"Invalid upload option value: {key}=\"{value}\"")

            # Handle deprecated options
            if base_key in list(self._OPTIONS_DEPRECATED.keys()):
                oldkey = key
                oldvalue = converted_value.value
                alt = key_part[1] + key_part[2]
                if base_key == "ignore-lintian":
                    key = f"lintian-mode{alt}"
                    converted_value.value = "ignore" if oldvalue else "errfail"
                elif base_key == "run-lintian":
                    key = f"lintian-mode{alt}"
                    converted_value.value = "errfail" if oldvalue else "disabled"
                LOG.warning(f"User Option: Converted DEPRECATED {oldkey}={oldvalue} to {key}={converted_value.value}")

            self._options[key] = converted_value.value

            LOG.debug(f"Upload option set: {key}=\"{value}\"")

        def get(self, key, alt=None, default=None):
            """Get first existing option value in this order: key[a], key, default."""
            # Validity check for key
            if key not in list(self._OPTIONS.keys()):
                raise mini_buildd.HTTPInternal(f"Unknown key used for get(): {key}")

            if alt:
                m_key = f"{key}[{alt}]"
                if m_key in self._options:
                    return self._options.get(m_key, default)
            return self._options.get(key, default)

    def __init__(self, file_path):
        super().__init__(file_path)
        if "Changed-By" not in self:
            raise mini_buildd.HTTPBadRequest("Changes has no 'Changed-By' field (maybe the changelog entry was not finalised?)")
        self.options = self.Options(self)

    @property
    def distribution(self):
        return mini_buildd.distribution.Distribution(mini_buildd.models.repository.map_distribution(self["distribution"]))

    def __str__(self):
        return f"User upload: {self.key}"

    def check_mode(self, checker, alt, fallback_mode_int):
        """Get check mode from upload options, with fallback."""
        options_mode_str = self.options.get(f"{checker}-mode", alt=alt)
        fallback_mode_str = mini_buildd.distribution.SbuildCheck.Mode(fallback_mode_int).name
        return mini_buildd.distribution.SbuildCheck(checker, options_mode_str if options_mode_str is not None else fallback_mode_str)

    def gen_buildrequests(self, repository, dist, suite_option):
        """
        Build buildrequest files for all architectures.

        .. todo:: **IDEA**: gen_buildrequests(): Instead of tar'ing ourselves (uploaded changes)
                  with exceptions (.deb, .buildinfo, .changes), add the *.dsc* and its files only!
        """
        # Extra check on all DSC/source package files
        # - Check md5 against possible pool files.
        # - Add missing from pool (i.e., orig.tar.gz).
        # - make sure all files from dsc are actually available
        files_from_pool = []
        with open(self.dsc_file_name) as dsc_file:
            dsc = debian.deb822.Dsc(dsc_file)

        for f in dsc["Files"]:
            in_changes = f["name"] in self.get_files(key="name")
            from_pool = False
            for p in glob.glob(os.path.join(repository.mbd_path, "pool", "*", "*", self["Source"], f["name"])):
                if f["md5sum"] == mini_buildd.misc.Hash(p).md5():
                    if not in_changes:
                        files_from_pool.append(p)
                        from_pool = True
                        LOG.info(f"Buildrequest: File added from pool: {p}")
                else:
                    raise mini_buildd.HTTPBadRequest(f"MD5 mismatch in uploaded dsc vs. pool: {f['name']}")

            # Check that this file is available
            if not in_changes and not from_pool:
                raise mini_buildd.HTTPBadRequest(f"Missing file '{f['name']}' neither in upload, nor in pool (use '-sa' for uploads with new upstream)")

        buildrequest_dict = {}
        for ao in dist.architectureoption_set.all():
            path = os.path.join(self.get_spool_dir(), ao.architecture.name)

            buildrequest = Buildrequest(os.path.join(path, self.gen_file_name(ao.architecture.name, Buildrequest.FILE_APX)))

            if buildrequest.is_new():
                import mini_buildd.models.repository  # pylint: disable=redefined-outer-name
                buildrequest["Distribution"] = self.distribution.get()
                for v in ["Source", "Version"]:
                    buildrequest[v] = self[v]

                # Generate files
                chroot_setup_script = os.path.join(path, "chroot_setup_script")
                with mini_buildd.open_utf8(os.path.join(path, "apt_sources.list"), "w") as asl, \
                     mini_buildd.open_utf8(os.path.join(path, "apt_preferences"), "w") as ap, \
                     mini_buildd.open_utf8(os.path.join(path, "apt_keys"), "w") as ak, \
                     mini_buildd.open_utf8(os.path.join(path, "ssl_cert"), "w") as ssl_cert, \
                     mini_buildd.open_utf8(chroot_setup_script, "w") as css, \
                     mini_buildd.open_utf8(os.path.join(path, "sbuildrc_snippet"), "w") as src:
                    asl.write(repository.mbd_get_apt_build_sources_list(dist, suite_option).get())
                    ap.write(repository.mbd_get_apt_build_preferences(dist, suite_option, self.options.get("internal-apt-priority")))
                    ak.write(repository.mbd_get_apt_keys(dist))
                    ssl_cert.write(mini_buildd.http_endpoint().get_certificate())
                    css.write(dist.chroot_setup_script)
                    os.chmod(chroot_setup_script, stat.S_IRWXU)
                    src.write(dist.sbuildrc_snippet)

                # Generate tar from original changes
                self.tar(tar_path=buildrequest.file_path + ".tar",
                         add_files=[os.path.join(path, "apt_sources.list"),
                                    os.path.join(path, "apt_preferences"),
                                    os.path.join(path, "apt_keys"),
                                    os.path.join(path, "ssl_cert"),
                                    chroot_setup_script,
                                    os.path.join(path, "sbuildrc_snippet")] + files_from_pool,
                         exclude_globs=["*.deb", "*.changes", "*.buildinfo"])
                buildrequest.add_file(buildrequest.file_path + ".tar")

                buildrequest.cset("Upload-To", mini_buildd.Daemon().model.mbd_get_ftp_endpoint().geturl())
                buildrequest.cset("Packager", mini_buildd.http_endpoint().geturl())
                buildrequest["Architecture"] = ao.architecture.name
                if ao.build_architecture_all:
                    buildrequest.cset("Arch-All", "Yes")
                buildrequest.cset("Build-Dep-Resolver", dist.get_build_dep_resolver_display())
                buildrequest.cset("Apt-Allow-Unauthenticated", "1" if dist.apt_allow_unauthenticated else "0")

                for blocks_option in mini_buildd.sbuild.CONFIG_BLOCKS.extra_options() + mini_buildd.sbuild.SETUP_BLOCKS.extra_options():
                    buildrequest.cset(blocks_option, dist.mbd_get_extra_option(blocks_option, ""))

                buildrequest.cset("Deb-Build-Options", self.options.get("deb-build-options", ao.architecture.name, dist.mbd_get_extra_option("Deb-Build-Options", "")))
                buildrequest.cset("Deb-Build-Profiles", self.options.get("deb-build-profiles", ao.architecture.name, dist.mbd_get_extra_option("Deb-Build-Profiles", "")))

                buildrequest.cset("Lintian-Mode", self.check_mode("lintian", ao.architecture.name, dist.lintian_mode).mode.name)
                buildrequest.cset("Lintian-Extra-Options", dist.lintian_extra_options)
                buildrequest.cset("Autopkgtest-Mode", self.check_mode("autopkgtest", ao.architecture.name, int(dist.mbd_get_extra_option("Autopkgtest-Mode", "0"))).mode.name)
                buildrequest.cset("Piuparts-Mode", self.check_mode("piuparts", ao.architecture.name, dist.piuparts_mode).mode.name)

                buildrequest.save(mini_buildd.Daemon().model.mbd_gnupg())
            else:
                LOG.info(f"Re-using existing buildrequest: {buildrequest.file_name}")
            buildrequest_dict[ao.architecture.name] = buildrequest

        return buildrequest_dict


class Buildrequest(Changes):  # pylint: disable=too-many-ancestors
    FILE_APX = "_mini-buildd-buildrequest"

    def __str__(self):
        return f"Buildrequest: {self.key} ({self['architecture']})"

    def check_mode(self, checker):
        """Get check mode from checker (lintian, piuparts, autopkgtest)."""
        return mini_buildd.distribution.SbuildCheck(checker, self.cget(f"{checker.capitalize()}-Mode", mini_buildd.distribution.SbuildCheck.Mode.disabled.name))

    def check_extra_options(self, checker):
        """Get check mode from checker (lintian, piuparts, autopkgtest)."""
        return self.cget(f"{checker.capitalize()}-Extra-Options", "")

    @property
    def live_buildlog_name(self):
        return f"{self.get_spool_id()}.buildlog"

    def get_live_buildlog_loc(self):
        return f"/mini_buildd/buildlogs/live/{self.live_buildlog_name}"

    def get_live_buildlog_url(self):
        return urllib.parse.urljoin(self.cget("Builder"), self.get_live_buildlog_loc())

    def upload_buildrequest(self, local_endpoint):
        import mini_buildd.models.gnupg  # pylint: disable=redefined-outer-name

        arch = self["Architecture"]
        codename = str(self.distribution.codename)

        builders = {}

        def add_builder(builder):
            status = builder.mbd_get_status()
            if status["running"] and mini_buildd.api.Status.has_chroot(status, codename, arch):
                builders[status["load"]] = status
                LOG.debug(f"Builder[{status['load']}]={builder}")

        # Always add our own instance as pseudo builder first.
        add_builder(mini_buildd.models.gnupg.Remote(http=local_endpoint.geturl()))

        # Check all active or auto-deactivated builders.
        for r in mini_buildd.models.gnupg.Remote.mbd_get_active_or_auto_reactivate():
            try:
                mini_buildd.models.gnupg.Remote.Admin.mbd_check(None, r, force=True)
                add_builder(r)
            except BaseException as e:
                mini_buildd.log_exception(LOG, "Builder check failed", e)

        if not builders:
            raise mini_buildd.HTTPBadRequest(f"No builder found for {codename}/{arch}")

        for _load, builder in sorted(builders.items()):
            try:
                self.cset("Builder", builder["http"])
                self.upload(mini_buildd.net.ClientEndpoint(builder["ftp"]))
                return
            except BaseException as e:
                mini_buildd.log_exception(LOG, f"Uploading to '{builder['ftp']}' failed", e)

        raise mini_buildd.HTTPBadRequest(f"Buildrequest upload failed for {arch}/{codename}")

    def gen_buildresult(self, path=None):
        if not path:
            path = self.get_spool_dir()

        buildresult = Buildresult(os.path.join(path, self.gen_file_name(self["Architecture"], Buildresult.FILE_APX)))

        for v in ["Distribution", "Source", "Version", "Architecture"]:
            buildresult[v] = self[v]

        buildresult.cset("Builder", mini_buildd.http_endpoint().geturl())
        buildresult.cset("Buildlog-Name", self.buildlog_name)
        buildresult.cset("Live-Buildlog", self.get_live_buildlog_url())

        return buildresult


class Buildresult(Changes):  # pylint: disable=too-many-ancestors
    FILE_APX = "_mini-buildd-buildresult"

    def __str__(self):
        return f"{self.key}: Sbuild={self.cget('Sbuild-Status')}, Lintian={self.cget('Sbuild-Lintian')}"


def select(file_path):
    if Buildresult.match(file_path):
        return Buildresult(file_path)
    if Buildrequest.match(file_path):
        return Buildrequest(file_path)
    return Upload(file_path)


class Log(LogPath):
    def __init__(self, distribution, source, version, create=False):
        super().__init__(distribution, source, version, create=create)
        self.events, self.upload, self.builds = [], None, {}

        if self.exists:
            for c in glob.glob(f"{self.fullpath}*.json"):
                self.events.append(mini_buildd.events.Event.load(c))

            for c in glob.glob(f"{self.fullpath}/**/*.changes", recursive=True):
                changes = select(c)
                if isinstance(changes, Upload):
                    self.upload = changes
                else:
                    arch = changes["architecture"]
                    self.builds.setdefault(arch, {})
                    self.builds[arch][type(changes).__name__.lower()] = changes
        elif not create:
            raise mini_buildd.HTTPBadRequest("No such log")

    def installed(self):
        return [event for event in self.events if event.type == mini_buildd.events.Type.INSTALLED]
