import os
import sys

import mini_buildd

#: This should never ever be changed
CHAR_ENCODING = "UTF-8"

#: Compute python-version dependent install path
PY_PACKAGE_PATH = f"/usr/lib/python{sys.version_info[0]}/dist-packages"

#: Global constant object to use when signalling shutdown via message queues
SHUTDOWN = "SHUTDOWN"

#: HTTPD log file names
LOG_FILE = "daemon.log"
ACCESS_LOG_FILE = "access.log"

#: (Debian package) path for the internal package templates
PACKAGE_TEMPLATES = "/usr/share/mini-buildd/package-templates"


class Uri:
    """URI string with some convenience functionality."""

    def __init__(self, uri):
        self.__uri = uri

    def get(self, *args, prefix=""):
        return os.path.join(prefix, self.__uri, *args)

    def getabs(self, *args):
        return self.get(*args, prefix="/")

    def url(self, *args, endpoint=None):
        ep = mini_buildd.http_endpoint() if endpoint is None else endpoint
        return ep.geturl(path=self.getabs(*args))


class Path:
    """PATH string with some convenience functionality."""

    def __init__(self, path, create=False):
        self.__path = path
        if len(self.__path) < 4:
            raise Exception(f"Possibly dangerous path '{self.__path}'")  # Be a bit paranoid since path.get() results are used to expire-purge things
        if create:
            os.makedirs(self.__path, exist_ok=True)

    def get(self, *args):
        return os.path.join(self.__path, *args)


class Resource:
    """Resource linked to a path with optional uris."""

    def __init__(self, path=None, create=False, uris=None):
        self.path = Path(path, create=create)
        self.uris = uris


URIS = {
    "events": {
        "static": Uri("events-log"),
        "view": Uri("mini_buildd/events-log"),
        "events": Uri("events"),
    },
    "api": {
        "api": Uri("mini_buildd/api"),
        "view": Uri("mini_buildd/api_view"),
    },
    "log": {
        "static": Uri("mini_buildd/log"),  # PRE20: 'package logs' were placed here
    },
    "buildlogs": {
        "view": Uri("mini_buildd/buildlogs"),
    },
    "login": {
        "view": Uri("accounts/login/"),
    },
    "home": {
        "view": Uri("mini_buildd/home.html"),
    },
    "static": {
        "static": Uri("static"),
    },
    "doc": {
        "static": Uri("doc"),
    },
    "repositories": {
        "static": Uri("repositories"),
    },
}

STATIC = Resource(path=f"{PY_PACKAGE_PATH}/mini_buildd/static", uris=URIS["static"])
DOC = Resource(path=os.path.realpath("/usr/share/doc/mini-buildd/html"), uris=URIS["doc"])


class Home(Resource, dict):
    def add(self, name, relpath, **kwargs):
        self[name] = Resource(path=os.path.join(self.path.get(), relpath), uris=URIS.get(name), create=True, **kwargs)

    def __init__(self, home):
        super().__init__(path=home, uris=URIS.get("home"))
        self.add("spool", "var/spool")
        self.add("tmp", "var/tmp")
        self.add("incoming", "incoming")
        self.add("repositories", "repositories")
        self.add("chroots", "var/chroots")
        self.add("chroots_lib", "var/chroots-libdir")
        self.add("events", "var/events-log")
        self.add("event_hooks", "etc/event-hooks")
        self.add("log", "var/log")


#: Variable items (see mini-buildd main script)
DEBUG = []
FOREGROUND = False
HOSTNAME = "localhost"
HTTP_ENDPOINTS = []
HOME = {}


def python37_systemcert_workaround():
    # - python 3.7/buster just does not use the system's default cert store, although documented as such.
    # - python 3.8/bullseye seems to work just fine.
    # - This seems to be due to openssl's set_default_verify_paths() (or maybe how it's called in py 3.7).
    # - Explicitly setting env "SSL_CERT_FILE" seems to fix openssl
    # - https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=805646 (may-be the issue).
    # I guess openssl is so secure just because it does not work 99% of the time :).
    if sys.version_info < (3, 8):
        os.environ["SSL_CERT_FILE"] = "/etc/ssl/certs/ca-certificates.crt"


python37_systemcert_workaround()
