import inspect
import enum
import random
import logging
import datetime
import calendar
import sched

import mini_buildd.misc
import mini_buildd.threads

LOG = logging.getLogger(__name__)


def _rdate(base_date, h_start, h_stop):
    return datetime.datetime(year=base_date.year, month=base_date.month, day=base_date.day,
                             hour=random.randrange(h_start, h_stop),
                             minute=random.randrange(60), second=random.randrange(60))


def next_yearly_run():
    """First day of next month between 0am and 1am."""
    today = datetime.date.today()
    return _rdate(datetime.datetime(year=today.year + 1, month=1, day=1), 0, 6)


def next_monthly_run():
    """First day of next month between 0am and 1am."""
    today = datetime.date.today()
    return _rdate(today.replace(day=1) + datetime.timedelta(calendar.monthrange(today.year, today.month)[1]), 0, 1)


def next_weekly_run():
    """Next sunday between 2am and 3am."""
    now = datetime.datetime.now()
    return _rdate(now + datetime.timedelta(days=7 - (now.isoweekday() % 7)), 2, 3)


def next_daily_run():
    """Next day between 4am and 6am."""
    return _rdate(datetime.datetime.now() + datetime.timedelta(days=1), 4, 6)


def next_minutely_run():
    """For testing only; start in one minute."""
    return datetime.datetime.now() + datetime.timedelta(minutes=1)


class Type(enum.Enum):
    YEARLY = enum.auto()
    MONTHLY = enum.auto()
    WEEKLY = enum.auto()
    DAILY = enum.auto()
    MINUTELY = enum.auto()


NEXT_RUN = {Type.YEARLY: next_yearly_run,
            Type.MONTHLY: next_monthly_run,
            Type.WEEKLY: next_weekly_run,
            Type.DAILY: next_daily_run,
            Type.MINUTELY: next_minutely_run}


class _Job():
    class Log():
        """Keep some data from job run for display (assure all times to be UTC)."""

        def __init__(self, job):
            self.job = job
            self.scheduled = job.next_run_utc
            self.error = None
            self.finished = None
            LOG.info(f"Starting cronjob '{job.desc}'...")

        def done(self):
            self.finished = mini_buildd.misc.datetime_stamp()
            (LOG.error if self.error else LOG.info)(f"Finished cronjob '{self.job.desc}' (error={self.error})")
            return self

        def seconds(self):
            return (self.finished - self.scheduled).total_seconds()

    def __enter(self):
        self.next_run = NEXT_RUN[self.type]()
        self.next_run_utc = self.next_run.astimezone(tz=datetime.timezone.utc)
        self.event = self.scheduler.enter((self.next_run - datetime.datetime.now()).total_seconds(), 1, self.__action, argument=self._argument, kwargs=self._kwargs)
        LOG.info(f"Scheduling cronjob '{self.desc}' for {self.next_run} (LOCAL), {self.next_run_utc} (UTC)")

    def __init__(self, scheduler, type_, action, argument=None, kwargs=None):
        self.scheduler = scheduler
        self.type = type_
        self._action = action
        self._argument = () if argument is None else argument
        self._kwargs = {} if kwargs is None else kwargs
        self.desc = inspect.getdoc(self._action)
        self.past_runs = []

        self.__enter()

    def __action(self, *args, **kwargs):
        log = _Job.Log(self)
        try:
            self._action(*args, **kwargs)
        except Exception as e:
            log.error = f"{e}"
        self.__enter()
        self.past_runs.append(log.done())


class Tab(mini_buildd.threads.PollerThread):
    def __init__(self):
        super().__init__(name="crontab")
        self.scheduler = sched.scheduler()
        self.jobs = []

    def add(self, type_, *args, **kwargs):
        self.jobs.append(_Job(self.scheduler, type_, *args, **kwargs))
        return self

    def run_poller(self):
        self.scheduler.run(blocking=False)
