import os
import shutil
import datetime
import threading
import glob
import queue
import collections
import urllib.request
import logging

import django.conf

import debian.deb822

import mini_buildd.config
import mini_buildd.misc
from mini_buildd.misc import Singleton  # Workaround for pylint crash. See https://github.com/PyCQA/pylint/issues/3742 and Daemon class below.
import mini_buildd.call
import mini_buildd.changes
import mini_buildd.gnupg
import mini_buildd.api
import mini_buildd.threads
import mini_buildd.ftpd
import mini_buildd.package
import mini_buildd.packager
import mini_buildd.builder
import mini_buildd.events
import mini_buildd.cron

import mini_buildd.models.daemon
import mini_buildd.models.repository
import mini_buildd.models.chroot
import mini_buildd.models.gnupg
import mini_buildd.models.subscription

LOG = logging.getLogger(__name__)


class Keyrings():
    """Hold/manage all gnupg keyrings (for remotes and all repository uploaders)."""

    def __init__(self, model):
        self._model = model
        self._our_pub_key = self._model.mbd_get_pub_key()
        self._remotes = self._gen_remotes()
        self._uploaders = self._gen_uploaders()
        self._needs_update = False

    def set_needs_update(self):
        self._needs_update = True

    def close(self):
        self._remotes.close()
        for u in list(self._uploaders.values()):
            u.close()

    def _update(self):
        if self._needs_update:
            self.close()
            self.__init__(self._model)

    def get_remotes(self):
        self._update()
        return self._remotes

    def get_uploaders(self):
        self._update()
        return self._uploaders

    def _gen_remotes(self):
        remotes = mini_buildd.gnupg.TmpGnuPG(tmpdir_options={"prefix": "gnupg-remotes-keyring-"})
        # keyring["remotes"]: Remotes keyring to authorize buildrequests and buildresults
        # Always add our own key
        if self._our_pub_key:
            remotes.add_pub_key(self._our_pub_key)
        for r in mini_buildd.models.gnupg.Remote.mbd_get_active_or_auto_reactivate():
            remotes.add_pub_key(r.key)
            LOG.info(f"Remote key added for '{r}': {r.key_long_id}: {r.key_name}")
        return remotes

    def _gen_uploaders(self):
        """All uploader keyrings for each repository."""
        uploaders = {}
        for r in mini_buildd.models.repository.Repository.mbd_get_active():
            uploaders[r.identity] = r.mbd_get_uploader_keyring()
            # Always add our key too for internal builds
            if self._our_pub_key:
                uploaders[r.identity].add_pub_key(self._our_pub_key)
        return uploaders


START_STOP_LOCK = threading.Lock()


class Daemon(mini_buildd.threads.EventThread, metaclass=Singleton):
    def __init__(self):
        self.model = mini_buildd.models.daemon.get()

        # Vars that are (re)generated when the daemon model is updated
        self.incoming_queue = queue.Queue()

        # p-d-registration uses DEFAULT_FROM_EMAIL django setting for sending emails; there is no other way to set this.
        # As this setting lives in the django "Daemon" instance, there is no way to set it before django is up.
        # This code updates a django setting at runtime: https://docs.djangoproject.com/en/1.7/topics/settings/#altering-settings-at-runtime
        # Some however actually may be changed: https://code.djangoproject.com/ticket/14628
        # We are confident that DEFAULT_FROM_EMAIL belongs to that group; also, this method is always called thread-locked
        django.conf.settings.DEFAULT_FROM_EMAIL = self.model.email_address

        self.keyrings = Keyrings(self.model)

        # Events
        events_maxlen = 3 * (self.model.show_last_packages + self.model.show_last_builds)
        try:
            self.events = mini_buildd.events.Events.from_json(self.model.mbd_get_json_data(), maxlen=events_maxlen)
        except BaseException as e:
            mini_buildd.log_exception(LOG, "Error reading persisted events (ignoring)", e, logging.WARN)
            self.events = mini_buildd.events.Events(maxlen=events_maxlen)

        self.builder = mini_buildd.builder.Builder(self.model.build_queue_size)   # TODO: adapt max parallel builds on_model_change ?
        self.packager = mini_buildd.packager.Packager()
        self.ftpd = mini_buildd.ftpd.FtpD(endpoint=self.model.mbd_get_ftp_endpoint(), queue=self.incoming_queue)
        self.crontab = mini_buildd.cron.Tab().add(mini_buildd.cron.Type.WEEKLY, self.cronjob_check) \
                                             .add(mini_buildd.cron.Type.WEEKLY, self.cronjob_expire_live_buildlogs) \
                                             .add(mini_buildd.cron.Type.WEEKLY, self.cronjob_expire_events_log) \
                                             .add(mini_buildd.cron.Type.WEEKLY, self.cronjob_snapshots, argument=[mini_buildd.cron.Type.WEEKLY]) \
                                             .add(mini_buildd.cron.Type.MONTHLY, self.cronjob_snapshots, argument=[mini_buildd.cron.Type.MONTHLY]) \
                                             .add(mini_buildd.cron.Type.YEARLY, self.cronjob_snapshots, argument=[mini_buildd.cron.Type.YEARLY])

        # self.crontab.add(mini_buildd.cron.Type.MINUTELY, self.cronjob_debug)  # Uncomment for crontab debugging only

        super().__init__(subthreads=[self.ftpd, self.packager, self.builder, self.crontab], events=self.incoming_queue, name="daemon")

    def __str__(self):
        return f"{'UP' if self.is_alive() else 'DOWN'}: {self.model}"

    def sync(self):
        self.model.refresh_from_db()
        self.model.mbd_set_json_data(self.events.to_json())
        self.model.save(update_fields=["pickled_data"], mbd_sync=False)
        self.ftpd.endpoint = self.model.mbd_get_ftp_endpoint()

    def join(self, timeout=None):
        super().join()
        self.events.shutdown()
        self.sync()

        # keyrings.close() is not called implicitly; this leaves tmp files around.
        # There should be a nicer way, really...
        try:
            self.keyrings.close()
        except BaseException:
            pass

    def run_event(self, event):
        if mini_buildd.changes.Buildrequest.match(event):
            self.builder.queue.put(event)
        else:
            self.packager.queue.put(event)  # User upload or build result: packager

    def cronjob_debug(self):
        """For debugging cron only."""
        import random
        if random.getrandbits(1):
            LOG.info(f"Running debug cron job for {self}")
        else:
            raise Exception(f"Failing randomly for {self}")

    def cronjob_check(self):
        """Check all instances."""
        mini_buildd.models.daemon.Daemon.Admin.mbd_action(None, (self.model,), "check")

    @classmethod
    def __fs_expire(cls, glob_pattern, older_than):
        """Expire events logs older than 1 year."""
        valid_until = mini_buildd.misc.datetime_stamp() - older_than
        LOG.debug(f"FS expire items from glob '{glob_pattern}' older than {older_than}...")
        for item in glob.glob(glob_pattern):
            if datetime.datetime.fromtimestamp(os.path.getmtime(item), tz=datetime.timezone.utc) < valid_until:
                LOG.info(f"Expiring FS item: {item} older than {older_than}...")
                if os.path.isdir(item):
                    shutil.rmtree(item)
                else:
                    os.remove(item)

    @classmethod
    def __fs_expire_empty_dirs(cls, path):
        LOG.debug(f"FS expire empty dirs under '{path}' ...")
        for root, dirs, _files in os.walk(path, topdown=False):
            for d in dirs:
                item = os.path.join(root, d)
                LOG.info(f"Expiring FS directory item (if empty): {item}")
                try:
                    os.rmdir(item)
                except OSError:
                    pass

    @classmethod
    def cronjob_expire_live_buildlogs(cls):
        """Expire live buildlogs older than 5 days."""
        cls.__fs_expire(mini_buildd.config.HOME["spool"].path.get("*.buildlog"), datetime.timedelta(days=5))

    @classmethod
    def cronjob_expire_events_log(cls):
        """Expire events log files older than 1 year."""
        cls.__fs_expire(mini_buildd.config.HOME["events"].path.get("*/*/*"), datetime.timedelta(weeks=4 * 12 * 1))  # Keep 1 year
        cls.__fs_expire_empty_dirs(mini_buildd.config.HOME["events"].path.get())

    def cronjob_snapshots(self, crontype):
        """Generate repository snapshots for all stable distributions."""
        prefix = f"{crontype.name}_"
        keep = {mini_buildd.cron.Type.YEARLY: datetime.timedelta(weeks=4 * 12 * 4),  # Keep  4 years
                mini_buildd.cron.Type.MONTHLY: datetime.timedelta(weeks=12 * 4),     # Keep 12 months
                mini_buildd.cron.Type.WEEKLY: datetime.timedelta(weeks=8),           # Keep  8 weeks: Max 24 automatic snapshots at a time
                mini_buildd.cron.Type.DAILY: datetime.timedelta(days=7),             # Keep 7 days (not in use)
                mini_buildd.cron.Type.MINUTELY: datetime.timedelta(minutes=5)}       # Keep 5 minutes (for TESTING only)

        for r in self.get_active_repositories():
            for d in r.distributions.all():
                for s in r.layout.suiteoption_set.filter(experimental=False, migrates_to=None):  # only for "stable" distributions
                    dist_str = r.mbd_get_distribution_string(d, s)
                    try:
                        r.mbd_reprepro.gen_snapshot(dist_str, f"{prefix}{datetime.datetime.utcnow().isoformat(timespec='minutes')}")
                    except Exception as e:  # on error, we should continue trying the other dists
                        LOG.error(f"{crontype.name} snapshot for {dist_str} failed: {e}")

                    try:
                        expire = datetime.datetime.utcnow().replace(second=0, microsecond=0) - keep[crontype]
                        LOG.info(f"Expiring all {crontype.name} snapshpots for {dist_str} older than {expire}")

                        for snap in r.mbd_reprepro.get_snapshots(dist_str, prefix=prefix):
                            stamp = datetime.datetime.fromisoformat(snap[len(prefix):])
                            if stamp <= expire:
                                r.mbd_reprepro.del_snapshot(dist_str, snap)
                            else:
                                LOG.info(f"KEEP: {snap}: {keep[crontype]}, {stamp}, {expire}")
                    except Exception as e:  # on error, we should continue trying the other dists
                        LOG.error(f"{crontype.name} expiration code for {dist_str} failed: {e}")

    @classmethod
    def get_status(cls):
        status = mini_buildd.api.Status()
        status.run()
        return status

    def get_title(self):
        """Human-readable short title for this Daemon instance."""
        return f"{self.model.identity}@{self.model.hostname}"

    @classmethod
    def meta(cls, model, func, request):
        model_class = eval(f"mini_buildd.models.{model}.Admin")  # pylint: disable=eval-used
        getattr(model_class, f"mbd_meta_{func}")(request)

    @classmethod
    def log_cat(cls, lines):
        with mini_buildd.open_utf8(mini_buildd.config.HOME["log"].path.get(mini_buildd.config.LOG_FILE)) as lf:
            return "".join(collections.deque(lf, lines))

    @classmethod
    def get_active_chroots(cls):
        return mini_buildd.models.chroot.Chroot.mbd_get_active()

    @classmethod
    def get_active_repositories(cls):
        return mini_buildd.models.repository.Repository.mbd_get_active()

    @classmethod
    def get_suites(cls):
        return mini_buildd.models.distribution.Suite.objects.all()

    @classmethod
    def get_active_codenames(cls):
        codenames = []
        for r in cls.get_active_repositories():
            for d in r.distributions.all():
                if d.base_source.codename not in codenames:
                    codenames.append(d.base_source.codename)
        return codenames

    @classmethod
    def get_active_or_auto_reactivate_remotes(cls):
        return mini_buildd.models.gnupg.Remote.mbd_get_active_or_auto_reactivate()

    @classmethod
    def get_subscription_objects(cls):
        return mini_buildd.models.subscription.Subscription.objects

    _DEFAULT_PORT_OPTIONS = ["lintian-mode=ignore"]

    def _port(self, dsc_url, package, dist, version, comments=None, options=None):
        def changelog_entries():
            """Merge comments and options into plain changelog_entries."""
            changelog_entries = [] if comments is None else comments
            for o in self._DEFAULT_PORT_OPTIONS if options is None else options:
                changelog_entries.append(f"{mini_buildd.changes.Upload.Options.KEYWORD}: {o}")
            return changelog_entries

        with mini_buildd.misc.TemporaryDirectory(prefix="port-") as tmpdir:
            env = mini_buildd.call.taint_env({"DEBEMAIL": self.model.email_address,
                                              "DEBFULLNAME": self.model.mbd_fullname(),
                                              "GNUPGHOME": self.model.mbd_gnupg().home})

            # Download DSC via dget.
            mini_buildd.call.Call(["dget",
                                   "--allow-unauthenticated",
                                   "--download-only",
                                   dsc_url],
                                  cwd=tmpdir,
                                  env=env).log().check()

            # Get SHA1 of original dsc file
            original_dsc_sha1sum = mini_buildd.misc.Hash(os.path.join(tmpdir, os.path.basename(dsc_url))).sha1()

            # Unpack DSC (note: dget does not support -x to a dedcicated dir).
            dst = "debian_source_tree"
            mini_buildd.call.Call(["dpkg-source",
                                   "-x",
                                   os.path.basename(dsc_url),
                                   dst],
                                  cwd=tmpdir,
                                  env=env).log().check()

            dst_path = os.path.join(tmpdir, dst)

            # Get version and author from original changelog; use the first block not
            with mini_buildd.open_utf8(os.path.join(dst_path, "debian", "changelog"), "r") as cl:
                original_author, original_version = mini_buildd.package.Changelog(cl, max_blocks=100).find_first_not(self.model.email_address)
            LOG.debug(f"Port: Found original version/author: {original_version}/{original_author}")

            # Change changelog in DST
            mini_buildd.call.Call(["debchange",
                                   "--newversion", version,
                                   "--force-distribution",
                                   "--force-bad-version",
                                   "--preserve",
                                   "--dist", dist,
                                   f"Automated port via mini-buildd (no changes). Original DSC's SHA1: {original_dsc_sha1sum}."],
                                  cwd=dst_path,
                                  env=env).log().check()

            for entry in changelog_entries():
                mini_buildd.call.Call(["debchange",
                                       "--append",
                                       entry],
                                      cwd=dst_path,
                                      env=env).log().check()

            # Repack DST
            mini_buildd.call.Call(["dpkg-source", "-b", dst], cwd=tmpdir, env=env).log().check()
            dsc = os.path.join(tmpdir,
                               mini_buildd.changes.Changes.gen_dsc_file_name(package,
                                                                             version))
            self.model.mbd_gnupg().sign(dsc)

            # Gen changes file name
            changes = os.path.join(tmpdir,
                                   mini_buildd.changes.Changes.gen_changes_file_name(package,
                                                                                     version,
                                                                                     "source"))

            # Generate Changes file
            with open(changes, "w+") as out:
                # Note: dpkg-genchanges has a home-brewed options parser. It does not allow, for example, "-v 1.2.3", only "-v1.2.3", so we need to use *one* sequence item for that.
                mini_buildd.call.Call(["dpkg-genchanges",
                                       "-S",
                                       "-sa",
                                       f"-v{original_version}",
                                       f"-D{mini_buildd.misc.CField('Originally-Changed-By').fullname}={original_author}"],
                                      cwd=dst_path,
                                      env=env,
                                      stdout=out).log().check()

            # Sign and upload
            self.model.mbd_gnupg().sign(changes)
            mini_buildd.changes.Changes(changes).upload(self.model.mbd_get_ftp_endpoint())
            return dist, package, version

    def port(self, package, from_dist, to_dist, version, options=None):
        # check from_dist
        from_repository, from_distribution, from_suite, _from_rollback = mini_buildd.models.repository.parse_distribution(from_dist)
        p = from_repository.mbd_package_find(package, distribution=from_dist, version=version)

        # check to_dist
        to_repository, to_distribution, to_suite, _to_rollback = mini_buildd.models.repository.parse_distribution(to_dist, check_uploadable=True)

        # Ponder version to use
        v = mini_buildd.package.DebianVersion(p["version"])
        if to_dist == from_dist:
            port_version = v.gen_internal_rebuild()
        else:
            port_version = v.gen_internal_port(from_repository.mbd_get_mandatory_version_regex(from_distribution, from_suite),
                                               to_repository.mbd_get_default_version(to_distribution, to_suite))

        _component, path = from_repository.mbd_get_dsc_path(from_distribution, package, p["version"])
        if not path:
            raise mini_buildd.HTTPBadRequest(f"Can't find DSC for {package}-{p['version']} in pool")

        return self._port(mini_buildd.http_endpoint().geturl(path=path), package, to_dist, port_version, options=options)

    def port_ext(self, dsc_url, to_dist, options=None):
        to_repository, to_distribution, to_suite, _to_rollback = mini_buildd.models.repository.parse_distribution(to_dist, check_uploadable=True)
        dsc = debian.deb822.Dsc(urllib.request.urlopen(dsc_url))
        v = mini_buildd.package.DebianVersion(dsc["Version"])
        return self._port(dsc_url,
                          dsc["Source"],
                          to_dist,
                          v.gen_external_port(to_repository.mbd_get_default_version(to_distribution, to_suite)),
                          comments=[f"External port from: {dsc_url}"],
                          options=options)

    def mbd_start(self):
        with START_STOP_LOCK:
            if not self.is_alive() and self.model.mbd_is_active():
                self.start()

    def mbd_stop(self):
        with START_STOP_LOCK:
            if self.is_alive():
                self.shutdown()
                self.join()
                Daemon.destroy()
