import enum
import re
import logging

import distro_info

import mini_buildd

LOG = logging.getLogger(__name__)


class SbuildCheck():
    """
    Generic support for sbuild checks (lintian, piuparts, autopkgtest).

    >>> SbuildCheck("lindian", "disabled")
    Traceback (most recent call last):
    ...
    mini_buildd.HTTPBadRequest: HTTP 400: Unknown sbuild checker: lindian (valid options: lintian,piuparts,autopkgtest)
    >>> SbuildCheck("lintian", "warnfall")
    Traceback (most recent call last):
    ...
    mini_buildd.HTTPBadRequest: HTTP 400: Unknown sbuild check mode: warnfall (valid options: disabled,ignore,errfail,warnfail)
    >>> sc = SbuildCheck("lintian", "warnfail")
    >>> sc.checker
    'lintian'
    >>> sc.mode
    <Mode.warnfail: 3>
    """

    class Mode(enum.Enum):
        disabled = 0
        ignore = 1
        errfail = 2
        warnfail = 3

        def __str__(self):
            return {
                self.disabled: "Don't run check",
                self.ignore: "Run check but ignore results",
                self.errfail: "Run check and fail on errors",
                self.warnfail: "Run check and fail on warnings",
            }[self]

    CHECKERS = ["lintian", "piuparts", "autopkgtest"]
    CHOICES = [(mode.value, mode.name) for mode in Mode]

    #: From sbuild source code: We may expect these textual statuses:
    STATUSES_PASS = ["pass", "info"]
    STATUSES_WARN = ["warn", "no tests"]
    STATUSES_FAIL = ["error", "fail"]

    def __init__(self, checker, mode):
        if checker not in self.CHECKERS:
            raise mini_buildd.HTTPBadRequest(f"Unknown sbuild checker: {checker} (valid options: {','.join(self.CHECKERS)})")
        self.checker = checker
        try:
            self.mode = self.Mode[mode]
        except KeyError:
            raise mini_buildd.HTTPBadRequest(f"Unknown sbuild check mode: {mode} (valid options: {','.join([mode.name for mode in self.Mode])})")

    @classmethod
    def desc(cls):
        return "Mode to control if a check should prevent package installation (for non-experimental suites)."

    @classmethod
    def usage(cls):
        return ", ".join([f"'{mode.name}' ({mode})" for mode in cls.Mode])

    def check(self, status, ignore=False):
        """Check if status is ok in this mode."""
        return \
            ignore or \
            status in self.STATUSES_PASS or \
            self.mode in [self.Mode.disabled, self.Mode.ignore] or \
            (status in self.STATUSES_WARN and self.mode is not self.Mode.warnfail)


class Codename():
    DEBIAN_INFO = distro_info.DebianDistroInfo()
    UBUNTU_INFO = distro_info.UbuntuDistroInfo()

    def __init__(self, codename):
        self.codename = codename

    def __str__(self):
        return self.codename

    def has_lintian_suppress(self):
        """Test if the distribution (identified by the codename) has a recent lintian with the '--suppress-tags' option."""
        return self.codename not in ["buzz", "rex", "bo", "hamm", "slink", "potato", "woody", "sarge", "etch", "lenny"]

    def lintian_options(self, check):
        result = []

        if self.has_lintian_suppress():
            result.append("--suppress-tags=bad-distribution-in-changes-file")

        result.append({
            check.Mode.disabled: "",
            check.Mode.ignore: "",
            check.Mode.errfail: "",
            check.Mode.warnfail: "--fail-on-warning ",
        }[check.mode])

        return " ".join(result)

    def produces_ddeb_appendix(self):
        """
        Get list of codenames that (oddly) use ddeb as file appendix.

        Some Ubuntu codenames produce 'ddeb' file appendix for automated debug packages.
        reprepro can't handle these yet -- so this is needed for a workaround for the wizard setup.

        See https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=730572
        """
        return self.codename in ["bionic", "cosmic", "disco", "eoan", "focal", "groovy"]

    def arch_optional(self, arch):
        """
        Ubuntu drops i386.

        'focal' still has i386 repo (debootrap still works), so i386 is still picked
        by mini-buildd's setup. However, builds no longer work (default resolver
        aptitude missing).
        Choosing "apt" resolver would master this hurdle, but this would also change
        the default resolver for amd64. Imho best solution for mini-buildd's setup
        is to make this arch optional.

        Ref: https://discourse.ubuntu.com/t/community-process-for-32-bit-compatibility/12598?u=d0od
        """
        optional = {"focal": ["i386"],
                    "groovy": ["i386"]}
        return arch in optional.get(self.codename, [])

    def apt_allow_unauthenticated(self):
        # jessie sources 'archive.debian.org' and 'ftp.debian.org' are not the same:
        # * jessie from archive.debian.org needs this setting (as the Release file is signed with an expired key, and jessie's apt just won't work).
        # * jessie from ftp.debian.org (was left there for LTS for 'amd64 armel armhf i386' only) would work w/o that flag, as it is signed differently.
        # However, both archives will match the jessie source, and both could potentially be used.
        options = {"jessie": True}
        return options.get(self.codename, False)


class Distribution():
    """
    A mini-buildd distribution string.

    Normal distribution:

    >>> d = Distribution("squeeze-test-stable")
    >>> d.codename.codename, d.repository, d.suite
    ('squeeze', 'test', 'stable')
    >>> d.get()
    'squeeze-test-stable'

    Rollback distribution:

    >>> d = Distribution("squeeze-test-stable-rollback5")
    >>> d.is_rollback
    True
    >>> d.codename.codename, d.repository, d.suite, d.rollback
    ('squeeze', 'test', 'stable', 'rollback5')
    >>> d.get()
    'squeeze-test-stable-rollback5'
    >>> d.rollback_no
    5

    Malformed distributions:

    >>> Distribution("-squeeze-stable")
    Traceback (most recent call last):
    ...
    mini_buildd.HTTPBadRequest: HTTP 400: Malformed distribution '-squeeze-stable': Must be '<codename>-<repoid>-<suite>[-rollback<n>]'

    >>> Distribution("squeeze--stable")
    Traceback (most recent call last):
    ...
    mini_buildd.HTTPBadRequest: HTTP 400: Malformed distribution 'squeeze--stable': Must be '<codename>-<repoid>-<suite>[-rollback<n>]'

    >>> Distribution("squeeze-test-stable-")
    Traceback (most recent call last):
    ...
    mini_buildd.HTTPBadRequest: HTTP 400: Malformed distribution 'squeeze-test-stable-': Must be '<codename>-<repoid>-<suite>[-rollback<n>]'

    >>> Distribution("squeeze-test-stable-rollback")
    Traceback (most recent call last):
    ...
    mini_buildd.HTTPBadRequest: HTTP 400: Malformed distribution 'squeeze-test-stable-rollback': Must be '<codename>-<repoid>-<suite>[-rollback<n>]'

    >>> Distribution("squeeze-test-stable-rolback0")
    Traceback (most recent call last):
    ...
    mini_buildd.HTTPBadRequest: HTTP 400: Malformed distribution 'squeeze-test-stable-rolback0': Must be '<codename>-<repoid>-<suite>[-rollback<n>]'
    """

    _REGEX = re.compile(r"^\w+-\w+-\w+?(-rollback\d+)?$")

    def __init__(self, distribution):
        if not self._REGEX.match(distribution):
            raise mini_buildd.HTTPBadRequest(f"Malformed distribution '{distribution}': Must be '<codename>-<repoid>-<suite>[-rollback<n>]'")

        self.distribution = distribution
        self._dsplit = self.distribution.split("-")
        self.codename = Codename(self._dsplit[0])
        self.repository = self._dsplit[1]
        self.suite = self._dsplit[2]
        self.is_rollback = len(self._dsplit) == 4
        self.rollback = self._dsplit[3] if self.is_rollback else None
        self.rollback_no = int(re.sub(r"\D", "", self.rollback)) if self.rollback else None

    def get(self, rollback=True):
        return "-".join(self._dsplit) if rollback else "-".join(self._dsplit[:3])


def guess_codeversion(release):
    """
    Guess the 'codeversion'.

    Aka the first two digits of a Debian release version; for releases
    without version, this falls back to the uppercase codename.

    In Debian,
      - point release <= sarge had the 'M.PrN' syntax (with 3.1 being a major release).
      - point release in squeeze used 'M.0.N' syntax.
      - point releases for >= wheezy have the 'M.N' syntax (with 7.1 being a point release).
      - testing and unstable do not gave a version in Release and fall back to uppercase codename

    Ubuntu just uses YY.MM which we can use as-is.

    >>> guess_codeversion({"Origin": "Debian", "Version": "3.1r8", "Codename": "sarge"})
    '31'
    >>> guess_codeversion({"Origin": "Debian", "Version": "4.0r9", "Codename": "etch"})
    '40'
    >>> guess_codeversion({"Origin": "Debian", "Version": "6.0.6", "Codename": "squeeze"})
    '60'
    >>> guess_codeversion({"Origin": "Debian", "Version": "7.0", "Codename": "wheezy"})
    '7'
    >>> guess_codeversion({"Origin": "Debian", "Version": "7.1", "Codename": "wheezy"})
    '7'
    >>> guess_codeversion({"Origin": "Debian", "Codename": "jessie"})
    '~JESSIE'
    >>> guess_codeversion({"Origin": "Debian", "Codename": "sid"})
    '~SID'
    >>> guess_codeversion({"Origin": "Ubuntu", "Version": "12.10", "Codename": "quantal"})
    '1210'

    """
    try:
        ver_split = release["Version"].split(".")
        number0 = ver_split[0]
        number1 = ver_split[1].partition("r")[0]  # Some older Debian versions had an "r" in the number
        if release.get("Origin", None) == "Debian" and int(number0) >= 7:
            return number0  # Debian >= wheezy: '~testN': One number tells the codeversion
        return number0 + number1  # Debian < wheezy, Ubuntu, maybe others: 1st plus 2nd number is needed
    except BaseException as e:
        LOG.info(f"Falling back to codename for codeversion due to: {e}")
        return "~" + release["Codename"].upper()
