import os.path
import enum
import queue
import datetime
import glob
import json
import threading
import collections
import weakref
import logging

import mini_buildd.config

import debian.debian_support

LOG = logging.getLogger(__name__)


class _NoLock():
    """
    Debug/temporary only: Experienced dead locks -- not reliably reproducible for now.

    May-be-related to our locking here. Use this to exclude our locking once we can
    reproduce the lockdown.
    """

    def acquire(self):
        pass

    def release(self):
        pass

    def __enter__(self):
        pass

    def __exit__(self, _type, _value, _traceback):
        pass


class Type(enum.Enum):
    REJECTED = enum.auto()
    BUILT = enum.auto()
    INSTALLED = enum.auto()
    FAILED = enum.auto()
    MIGRATED = enum.auto()
    REMOVED = enum.auto()


class Event():
    def __init__(self, type_, distribution, source, version, extra):
        self.type = type_
        self.distribution = distribution
        self.source = source
        self.version = version
        self.timestamp = mini_buildd.misc.datetime_stamp()
        self.extra = extra

    def extra_info(self):
        """Human-readable informational string from extra data."""
        try:
            if self.type in [Type.REJECTED, Type.FAILED]:
                error = self.extra.get("error")
                return f"{mini_buildd.Rfc7807.from_json(error)}" if error else ""
            if self.type in [Type.BUILT]:
                return f"{self.extra.get('architecture')} {self.extra.get('sbuild_status')} (builder {self.extra.get('builder')})"
            if self.type in [Type.INSTALLED]:
                return ', '.join([f"{arch} {bres.get('sbuild_status')}" for arch, bres in self.extra.get("buildresults").items()])
        except BaseException as e:
            mini_buildd.log_exception(LOG, "event extra_info() display failed", e)
        return ""

    def __str__(self):
        return ": ".join(filter(None, [f"{self.type.name}({self.distribution}): {self.source}-{self.version}", self.extra_info()]))

    def match(self, types=None, distribution=None, source=None, version=None, minimal_version=None):
        def match(match, value):
            return match is None or match == value

        # pylint: disable=too-many-boolean-expressions
        return \
            (types is None or self.type in types) and \
            match(distribution, self.distribution) and \
            match(source, self.source) and \
            match(version, self.version) and \
            (minimal_version is None or debian.debian_support.Version(self.version) >= debian.debian_support.Version(minimal_version))

    def to_json(self):
        return {
            "type": self.type.name,
            "distribution": self.distribution,
            "source": self.source,
            "version": self.version,
            "timestamp": self.timestamp.isoformat(),
            "extra": self.extra,
        }

    @classmethod
    def from_json(cls, data):
        event = Event(Type[data["type"]],
                      data["distribution"],
                      data["source"],
                      data["version"],
                      data["extra"])
        event.timestamp = datetime.datetime.fromisoformat(data["timestamp"])
        return event

    @classmethod
    def from_changes(cls, typ, changes, extra=None):
        return Event(typ,
                     distribution=changes.get("distribution"),
                     source=changes.get("source"),
                     version=changes.get("version"),
                     extra={**{"architecture": changes.get("architecture"),
                               "maintainer": changes.get("maintainer"),
                               "changed_by": changes.get(mini_buildd.misc.CField("originally-changed-by").fullname, changes.get("changed-by")),
                               },
                            **(extra or {}),
                            })

    def log_path(self, create=False):
        return mini_buildd.changes.LogPath(self.distribution, self.source, self.version, create=create)

    def save(self, file_path=None):
        if file_path is None:
            arch_apx = f"_{self.extra['architecture']}" if self.type == Type.BUILT else ""  # Only BUILT type is ambiguous w/o arch.
            file_path = self.log_path(create=True).fullpath + f".{self.type.name}{arch_apx}.json"

        with open(file_path, "w") as f:
            json.dump(self.to_json(), f)

        hook_pattern = "[!_.]*[!~]"
        for hook in sorted(glob.glob(mini_buildd.config.HOME["event_hooks"].path.get(hook_pattern))):
            if os.path.isfile(hook) and os.access(hook, os.X_OK):
                try:
                    mini_buildd.call.Call([hook, file_path]).log().check()
                except BaseException as e:
                    LOG.warning(f"Custom Event Hook failed: {hook}: {e}")

    @classmethod
    def load(cls, file_path):
        with open(file_path) as f:
            return Event.from_json(json.load(f))


class Events(collections.deque):
    MAX_CLIENTS = 100

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._clients = {}
        self._lock = threading.Lock()

    def log(self, event):
        LOG.debug(f"Logging event: {event}")
        event.save()
        with self._lock:
            self.appendleft(event)
            mini_buildd.Daemon().model.mbd_notify_event(event)

            for _key, (weakref_obj, q) in self._clients.items():
                if weakref_obj() is not None:
                    LOG.info(f"Queuing event for object '{weakref_obj}'")
                    q.put(event)

    def logchanges(self, typ, changes, extra=None):
        self.log(Event.from_changes(typ, changes, extra))

    def attach(self, obj, since=None):
        with self._lock:
            # Garbage collection
            for key in [key for key, (weakref_obj, _q) in self._clients.items() if weakref_obj() is None]:
                LOG.info(f"Detach event queue from object '{key}': Dead.")
                del self._clients[key]

            if id(obj) not in self._clients:
                if len(self._clients) >= Events.MAX_CLIENTS:
                    raise mini_buildd.HTTPUnavailable(f"Max client limit ({Events.MAX_CLIENTS}) exceeded")
                LOG.info(f"Attaching new event queue to object: {obj}")
                self._clients[id(obj)] = (weakref.ref(obj), queue.Queue())
                if since:
                    for event in reversed(self):
                        if event.timestamp >= since:
                            self._clients[id(obj)][1].put(event)

        return self._clients[id(obj)][1]

    def shutdown(self):
        """Hint shutdown to all client (queues). Essentially makes blocking get() in httpd.py continue so httpd can shutdown."""
        with self._lock:
            for client in self._clients.values():
                try:
                    client[1].put_nowait(mini_buildd.config.SHUTDOWN)
                except Exception as e:
                    LOG.info(f"Events: Ignoring failure to hint shutdown to client {client[0]}: {e}")

    def get_last_sources(self):
        return sorted({event.source for event in self if event.source})

    def get_last_versions(self, source):
        return sorted({event.version for event in self if event.source == source})

    def to_json(self):
        return [event.to_json() for event in self]

    @classmethod
    def from_json(cls, events, maxlen):
        return Events([Event.from_json(event) for event in events], maxlen=maxlen)


class Attach():
    def __init__(self, events, *args, **kwargs):
        self.events = events.attach(self, *args, **kwargs)

    def get(self):
        return self.events.get()


def filter_iter(events_iter, types=None, distribution=None, source=None, version=None, minimal_version=None, exit_on=None, fail_on=None):
    for event in events_iter():
        if event.match(types=types, distribution=distribution, source=source, version=version, minimal_version=minimal_version):
            yield event
            if fail_on is not None and event.type in fail_on:
                raise Exception(f"Failed on event: {event}")
            if exit_on is not None and event.type in exit_on:
                break
