import re
import os.path
import json
import datetime
import logging

import twisted.internet.endpoints
import twisted.internet.threads
import twisted.web.wsgi
import twisted.web.static
import twisted.web.resource
import twisted.web.server
import twisted.logger
import twisted.python.logfile

import mini_buildd.config
import mini_buildd.threads

LOG = logging.getLogger(__name__)


class Site(twisted.web.server.Site):
    def _openLogFile(self, path):  # noqa (pep8 N802)
        return twisted.python.logfile.LogFile(os.path.basename(path), directory=os.path.dirname(path), rotateLength=5000000, maxRotatedFiles=9)


class RootResource(twisted.web.resource.Resource):
    """Twisted root resource needed to mix static and wsgi resources."""

    def __init__(self, wsgi_resource):
        super().__init__()
        self._wsgi_resource = wsgi_resource

    def getChild(self, path, request):  # noqa (pep8 N802)
        request.prepath.pop()
        request.postpath.insert(0, path)
        return self._wsgi_resource


class FileResource(twisted.web.static.File):
    """Twisted static resource enhanced with switchable index and regex matching support."""

    def __init__(self, *args, with_index=False, uri_regex=".*", with_cache=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.defaultType = f"text/plain; charset={mini_buildd.config.CHAR_ENCODING}"

        self.mbd_with_index = with_index
        self.mbd_uri_regex = re.compile(uri_regex)
        self.mbd_with_cache = with_cache

    def directoryListing(self):  # noqa (pep8 N802)
        if not self.mbd_with_index:
            return self.forbidden
        return super().directoryListing()

    def getChild(self, path, request):  # noqa (pep8 N802)
        """Custom getChild implementation."""
        if not self.mbd_uri_regex.match(request.uri.decode("utf-8")):
            return self.forbidden
        child = super().getChild(path, request)
        child.mbd_with_index = self.mbd_with_index
        child.mbd_uri_regex = self.mbd_uri_regex
        child.mbd_with_cache = self.mbd_with_cache
        return child

    def render(self, request):
        if self.mbd_with_cache:
            request.setHeader("Cache-Control", "public, max-age=86400, no-transform")
        return super().render(request)


class Events(twisted.web.resource.Resource):
    @classmethod
    def _render(cls, request):
        try:
            LOG.debug(f"{request.channel}: Waiting for event...")
            since = datetime.datetime.fromisoformat(request.args[b"since"][0].decode(mini_buildd.config.CHAR_ENCODING)) if b"since" in request.args else None
            queue = mini_buildd.Daemon().events.attach(request.channel, since=mini_buildd.misc.datetime_check_aware(since))
            event = queue.get()
            LOG.debug(f"{request.channel}: Got event: {event}")
            if event is mini_buildd.config.SHUTDOWN:
                raise mini_buildd.HTTPShutdown
            request.write(json.dumps(event.to_json()).encode(mini_buildd.config.CHAR_ENCODING))
            request.finish()
        except Exception as e:
            http_exception = mini_buildd.e2http(e)
            request.setResponseCode(500, f"{http_exception}".encode(mini_buildd.config.CHAR_ENCODING))
            if getattr(request, "_disconnected", False):
                request.notifyFinish()
            else:
                request.finish()

    def render_GET(self, request):  # pylint: disable=invalid-name  # noqa (pep8 N802)
        request.setHeader("Content-Type", f"application/json; charset={mini_buildd.config.CHAR_ENCODING}")
        twisted.internet.threads.deferToThread(self._render, request)
        return twisted.web.server.NOT_DONE_YET


class HttpD(mini_buildd.threads.Thread):
    DOC_MISSING_HTML = ("<html><body>"
                        "<h1>Online Manual Not Available (<code>mini-buildd-doc</code> not installed?)</h1>"
                        "Package <b><code>mini-buildd-doc</code></b> is not installed on this site (might be intentional to save space)."
                        "<p><a href='/'>[Back]</a></p>"
                        "</body></html>")

    def _add_static_route(self, resource, with_index=False, uri_regex=".*", with_doc_missing_error=False, with_cache=False):
        static = FileResource(with_index=with_index, uri_regex=uri_regex, with_cache=with_cache, path=resource.path.get())

        if with_doc_missing_error:
            static.childNotFound = twisted.web.resource.NoResource(self.DOC_MISSING_HTML)

        self.resource.putChild(bytes(resource.uris["static"].get(), encoding=mini_buildd.config.CHAR_ENCODING), static)

    def __init__(self, wsgi_app):
        super().__init__(name="httpd")

        # Bend twisted (not access.log) logging to ours
        twisted.logger.globalLogPublisher.addObserver(twisted.logger.STDLibLogObserver(name=__name__))

        # HTTP setup
        self.resource = RootResource(twisted.web.wsgi.WSGIResource(mini_buildd.net.reactor(), mini_buildd.net.reactor().getThreadPool(), wsgi_app))  # pylint: disable=no-member
        self.site = Site(self.resource, logPath=mini_buildd.config.HOME["log"].path.get(mini_buildd.config.ACCESS_LOG_FILE))

        for ep in mini_buildd.config.HTTP_ENDPOINTS:
            twisted.internet.endpoints.serverFromString(mini_buildd.net.reactor(), ep.description).listen(self.site)

        # Note (with_cache=False): If browser cache should be enabled at some point, then here. However, we have no "hashed-on-change" support and it's confusing if
        # 'though F5, js still running old code'. Also a real PITA in most browsers to clear the cache.
        self._add_static_route(mini_buildd.config.STATIC,
                               with_cache=False)
        self._add_static_route(mini_buildd.config.DOC,
                               with_doc_missing_error=True)
        self._add_static_route(mini_buildd.config.HOME["repositories"],
                               with_index=True,
                               uri_regex=rf"^/{mini_buildd.config.HOME['repositories'].uris['static'].get()}/.+/(pool|dists)/.*")
        self._add_static_route(mini_buildd.config.HOME["events"],
                               with_index=True)
        self._add_static_route(mini_buildd.config.HOME["log"],
                               with_index=True,
                               uri_regex=r"^/log/.+/.*")

        # Events route
        self.resource.putChild(bytes(mini_buildd.config.URIS["events"]["events"].get(), encoding=mini_buildd.config.CHAR_ENCODING), Events())

    def shutdown(self):
        mini_buildd.net.reactor().stop()  # pylint: disable=no-member

    def run(self):
        mini_buildd.net.reactor().run(installSignalHandlers=0)  # pylint: disable=no-member
        LOG.debug(f"{self}: FINISHED")
