import re
import socket
import email.utils
import inspect
import html
import json
import logging

import django.db
import django.core.mail
import django.core.exceptions

import mini_buildd.misc
import mini_buildd.net
import mini_buildd.gnupg

import mini_buildd.models.base
import mini_buildd.models.repository
import mini_buildd.models.chroot
import mini_buildd.models.gnupg

from mini_buildd.models.msglog import MsgLog
LOG = logging.getLogger(__name__)


class Daemon(mini_buildd.models.base.StatusModel):
    # Basics
    identity = django.db.models.CharField(max_length=50, default=socket.gethostname(),
                                          help_text=("<p>Daemon's identity; this will be used to identify this "
                                                     "mini-buildd instance in various places.</p>"
                                                     "<p>It will occur in the 'Name-Real' part of the GnuPG key, "
                                                     "determines the name of the automated keyring packages, and will "
                                                     "also occur in the 'Origin' tag of repository indices.</p>"
                                                     "<p>In case you don't like the automatic archive 'Origin' string, you "
                                                     "can customize it via 'X-Archive-Origin' extra option (see 'Extra options' below).</p>"))

    hostname = django.db.models.CharField(
        max_length=200,
        default=socket.getfqdn(),
        help_text="Obsoleted: Use command line arg '--hostname' instead.")

    email_address = django.db.models.EmailField(max_length=255, default=f"mini-buildd@{socket.getfqdn()}",
                                                help_text=("Daemon's email address; this will be used to identify this\n"
                                                           "mini-buildd instance in various places.\n"
                                                           "\n"
                                                           "It will occur in the 'Name-Email' part of the GnuPG key, and used as\n"
                                                           "maintainer email on various occations when packages are build\n"
                                                           "automatically.\n"))

    # GnuPG options
    gnupg_template = django.db.models.TextField(default=("Key-Type: RSA\n"
                                                         "Key-Length: 4096\n"
                                                         "Expire-Date: 0\n"),
                                                help_text=("Template as accepted by 'gpg --batch --gen-key' (see 'man gpg').\n"
                                                           "\n"
                                                           "You should not give 'Name-Real' and 'Name-Email', as these are\n"
                                                           "automatically added.)\n"))

    gnupg_keyserver = django.db.models.CharField(
        max_length=200,
        default="pool.sks-keyservers.net",
        help_text="GnuPG key server to use (as fill-in helper).")

    ftpd_bind = django.db.models.CharField(
        max_length=200,
        default="tcp:port=8067",
        help_text=f"FTP Server Endpoint: <pre>{html.escape(inspect.getdoc(mini_buildd.net.ServerEndpoint))}</pre>")

    ftpd_options = django.db.models.CharField(max_length=255, default="", blank=True, help_text="For future use.")

    # Load options
    build_queue_size = django.db.models.IntegerField(
        default=mini_buildd.misc.get_cpus(),
        help_text="Maximum number of parallel builds.")

    sbuild_jobs = django.db.models.IntegerField(
        default=1,
        help_text=("DEPRECATED/UNUSED (since 1.99.17): Degree of parallelism per build (via sbuild's '--jobs' option)."
                   "<pre>"
                   "No longer used because: \n"
                   " * The --jobs options of sbuild (resp. dpkg-buildpackage) forces builds to be parallel (if N > 1), possible leading to unfixable build problems.\n"
                   " * The --jobs options of sbuild (resp. dpkg-buildpackage) forces builds to be non-parallel (if N = 1, the default). This is not what you want for most packages.\n"
                   " * For some time now, DEB_BUILD_OPTIONS is set to parallel=<CPUS> by default by dpkg tools. Packages may act on this, and build parallel. Practically, most "
                   "packages will also just do so by default, when using build tools like debhelper.\n"
                   " * I.e., we now go completely with the defaults now. Use s.th. like 'dh --no-parallel' in your rules if your package does not support parallel building.\n"
                   "</pre>"
                   ))

    # EMail options
    # DEPRECATED/UNUSED: With the switch to django mail framework, this is now configured via the --smtp command line argument.
    smtp_server = django.db.models.CharField(
        max_length=254,
        default="localhost:25",
        help_text="DEPRECATED/UNUSED: Replaced by '--smtp' command line option.")

    notify = django.db.models.ManyToManyField(mini_buildd.models.repository.EmailAddress,
                                              blank=True,
                                              help_text="Addresses that get all notification emails unconditionally.")
    allow_emails_to = django.db.models.CharField(
        max_length=254,
        default=f".*@{socket.getfqdn()}",
        help_text=("Regex to allow sending emails to automatically computed\n"
                   "addresses (currently the 'Changed-By' or 'Maintainer' addresses\n"
                   "when this feature is enabled for the resp. repository).\n"
                   "\n"
                   "Use '.*' to allow all -- it's however recommended to put this to\n"
                   "s.th. like '.*@myemail.domain', to avoid original package\n"
                   "maintainers to be accidentally spammed.\n"))

    custom_hooks_directory = django.db.models.CharField(max_length=255, default="", blank=True, help_text="For future use.")

    show_last_packages = django.db.models.IntegerField(
        default=100,
        help_text="How many packager events to keep (2.0: this value will just be added to the max size of the generic event log).")
    show_last_builds = django.db.models.IntegerField(
        default=100,
        help_text="How many builder events to keep (2.0: this value will just be added to the max size of the generic event log).")

    wait_for_build_results = django.db.models.IntegerField(
        default=5,
        help_text="Future use: How many days to wait for build results until finishing a package.")
    keep_build_results = django.db.models.IntegerField(
        default=5,
        help_text="Future use: How many days to keep build results that cannot be uploaded.")

    LETHAL_DEPENDENCIES = False

    class Meta(mini_buildd.models.base.StatusModel.Meta):
        verbose_name_plural = "Daemon"

    class Admin(mini_buildd.models.base.StatusModel.Admin):
        fieldsets = (
            (None, {"fields": (), "description": ("The daemon instance. There is always exactly one instance of this.\n"
                                                  "\n"
                                                  "prepare/remove actions will generate/remove the GnuPG key.\n")}),
            ("Archive identity", {"fields": (("identity", "hostname", "email_address"), "gnupg_template")}),
            ("FTP (incoming) Options", {"fields": ("ftpd_bind", "ftpd_options")}),
            ("Load Options", {"fields": ("build_queue_size", "sbuild_jobs")}),
            ("E-Mail Options", {"fields": ("smtp_server", "notify", "allow_emails_to")}),
            ("Other Options", {"fields": ("gnupg_keyserver", "custom_hooks_directory", "show_last_packages", "show_last_builds", "extra_options")}))

        filter_horizontal = ("notify",)

        # These are depcrecated or not used yet
        readonly_fields = ["sbuild_jobs", "smtp_server", "ftpd_options", "custom_hooks_directory"]

        @classmethod
        def _mbd_on_activation(cls, request, obj):
            mini_buildd.Daemon().mbd_start()

        @classmethod
        def _mbd_on_deactivation(cls, request, obj):
            mini_buildd.Daemon().mbd_stop()

    def save(self, *args, mbd_sync=True, **kwargs):  # pylint: disable=arguments-differ
        """Save model and optionally sync with Daemon()."""
        super().save(*args, **kwargs)
        if mbd_sync:
            mini_buildd.Daemon().sync()

    def __str__(self):
        return (f"{self.identity}: Serving {len(mini_buildd.models.repository.Repository.mbd_get_active())} repositories, "
                f"{len(mini_buildd.models.chroot.Chroot.mbd_get_active())} chroots, "
                f"using {len(mini_buildd.models.gnupg.Remote.mbd_get_active())} remotes")

    def mbd_get_ftp_endpoint(self):
        return mini_buildd.net.ServerEndpoint(self.ftpd_bind, protocol="ftp")

    def mbd_fullname(self):
        return f"mini-buildd archive {self.identity}"

    def mbd_gnupg(self):
        return mini_buildd.gnupg.GnuPG(self.gnupg_template, self.mbd_fullname(), self.email_address)

    # Note: pylint false-positive: https://github.com/PyCQA/pylint/issues/1553
    def clean(self, *args, **kwargs):  # pylint: disable=arguments-differ
        super().clean(*args, **kwargs)
        self.mbd_validate_regex(r"^[a-zA-Z0-9\-]+$", self.identity, "identity")
        if Daemon.objects.count() > 0 and self.id != Daemon.objects.get().id:
            raise django.core.exceptions.ValidationError("You can only create one Daemon instance!")

    def mbd_prepare(self, request):
        self.mbd_gnupg().prepare()
        MsgLog(LOG, request).info("Daemon GnuPG prepared.")

    @classmethod
    def mbd_sync(cls, request):
        MsgLog(LOG, request).warning("The GnuPG key will never be updated automatically. Explicitly run remove+prepare to achieve this.")

    def mbd_remove(self, request):
        self.mbd_gnupg().remove()
        MsgLog(LOG, request).info("Daemon GnuPG removed.")

    def mbd_get_dependencies(self):
        """All active or to-be active repositories, remotes and chroots."""
        result = []
        for model in [mini_buildd.models.repository.Repository, mini_buildd.models.chroot.Chroot, mini_buildd.models.gnupg.Remote]:
            for o in model.mbd_get_active_or_auto_reactivate():
                result.append(o)
        return result

    def mbd_check(self, request):  # pylint: disable=no-self-use
        """Just warn in case there are no repos and no chroots."""
        if not mini_buildd.models.repository.Repository.mbd_get_active() and not mini_buildd.models.chroot.Chroot.mbd_get_active():
            MsgLog(LOG, request).warning("No active chroot or repository.")

    def mbd_get_archive_origin(self):
        return self.mbd_get_extra_option("X-Archive-Origin", f"Mini-Buildd {self.identity}")

    def mbd_get_pub_key(self):
        return self.mbd_gnupg().get_pub_key()

    def mbd_get_dput_conf(self):
        ftp = self.mbd_get_ftp_endpoint()
        return f"""\
[mini-buildd-{self.identity}]
method   = {'ftps' if ftp.is_ssl() else 'ftp'}
fqdn     = {ftp.hopo()}
login    = anonymous
incoming = /incoming
"""

    def _mbd_notify_signature(self, typ):
        reason = {"daemon": "Your address is configured to get any notifications (contact administrators if you don't want this).",
                  "repository": "Your address is configured to get any notifications for this repository (contact administrators if you don't want this).",
                  "changed-by": "Your address is the uploader of the package ('Changed-By' in changes).",
                  "maintainer": "Your address is the maintainer of the package ('Maintainer' in changes).",
                  "subscriber": "Your user account has a matching subscription.",
                  }.get(typ, "Unknown")

        return (f"--\N{SPACE}\n"
                f"mini-buildd instance '{self.identity}' at {mini_buildd.config.HOSTNAME} <{self.email_address}>\n"
                f"Reason for this mail: {reason}\n"
                f"Visit mini-buildd   : {mini_buildd.http_endpoint().geturl()}\n"
                f"Manage your account : {mini_buildd.config.URIS['login']['view'].url()}\n")

    def mbd_notify_event(self, event):
        repository, distribution = None, None
        try:
            repository, distribution, _suite, _rollback_no = mini_buildd.models.repository.parse_distribution(event.distribution)
        except BaseException:
            pass

        try:
            subject = f"{event}"
            body = json.dumps(event.to_json(), indent=2) + "\n"

            # Old mbd_notify() code:
            subject_prefix = f"[mini-buildd-{self.identity}] "
            m_to = []
            m_to_raw = []
            m_automatic_to_allow = re.compile(self.allow_emails_to)

            def add_to(address, typ, is_automatic):
                address_raw = email.utils.parseaddr(address)[1]
                if not is_automatic or m_automatic_to_allow.search(address_raw):
                    if address_raw in m_to_raw:
                        LOG.debug(f"Notify: Skipping {typ} address: {address}: Duplicate")
                    else:
                        m_to.append((subject_prefix + subject, body + self._mbd_notify_signature(typ), self.email_address, [address]))
                        m_to_raw.append(address_raw)
                        LOG.info(f"Notify: Adding {typ} address: {address}")
                else:
                    LOG.warning(f"Notify: Skipping {typ} address: {address}: Not allowed (only '{self.allow_emails_to}')")

            def get_subscriptions():
                source = "" if event.source is None else event.source
                real_distribution = distribution
                if real_distribution is None:
                    # If distribution was not given explicitly, try from changes, resolving meta dists if needed.
                    changes_dist = "" if event.distribution is None else event.distribution
                    real_distribution = mini_buildd.models.repository.map_distribution(changes_dist)
                return mini_buildd.models.subscription.Subscription.objects.filter(package__in=[source, ""], distribution__in=[real_distribution, ""])

            # Add hardcoded addresses from daemon
            for m in self.notify.all():
                add_to(m.__str__(), "daemon", is_automatic=False)

            # Add hardcoded addresses from repository
            if repository:
                for m in repository.notify.all():
                    add_to(m.__str__(), "repository", is_automatic=False)

            # Add package uploader (Changed-By): Add even if we do not have a repo, so uploader is informed on such error cases too.
            if event.extra.get("changed_by") and (repository is None or repository.notify_changed_by):
                add_to(event.extra.get("changed_by"), "changed-by", is_automatic=True)

            # Add package maintainer: Add only when we have a repo, and it's configured to do so
            if event.extra.get("maintainer") and repository and repository.notify_maintainer:
                add_to(event.extra.get("maintainer"), "maintainer", is_automatic=True)

            # Add user subscriptions
            for s in get_subscriptions():
                address = f"{s.subscriber.get_full_name()} <{s.subscriber.email}>"
                if s.subscriber.is_active:
                    add_to(address, "subscriber", is_automatic=False)
                else:
                    LOG.debug(f"Notify: Skipping subscription address: {address}: Account disabled")

            django.core.mail.send_mass_mail(m_to)
            LOG.info(f"EMail notification for event ({event}) sent.")
        except BaseException as e:
            mini_buildd.log_exception(LOG, f"EMail notification for event ({event}) FAILED", e)


def get():
    model, created = mini_buildd.models.daemon.Daemon.objects.get_or_create(id=1)
    if created:
        LOG.info("New Daemon model instance created")
    return model
