import tempfile
import urllib.request
import urllib.error
import logging
import contextlib
import functools
import copy
import socket

import dateutil.parser

import django.db.models
import django.utils.timezone
import django.core.exceptions

import debian.deb822

import mini_buildd.distribution
import mini_buildd.files
import mini_buildd.net
import mini_buildd.call
import mini_buildd.gnupg

import mini_buildd.models.base
import mini_buildd.models.gnupg

from mini_buildd.models.msglog import MsgLog
LOG = logging.getLogger(__name__)


class Archive(mini_buildd.models.base.Model):
    URLOPEN_TIMEOUT = 15  # seconds

    url = django.db.models.URLField(primary_key=True, max_length=512,
                                    default="http://ftp.debian.org/debian/",
                                    help_text=("The URL of an apt archive (there must be a 'dists/' infrastructure below).\n"
                                               "\n"
                                               "Use the 'directory' notation with exactly one trailing slash (like 'http://example.org/path/').\n"))
    ping = django.db.models.FloatField(default=-1.0, editable=False)

    class Meta(mini_buildd.models.base.Model.Meta):
        ordering = ["url"]

    class Admin(mini_buildd.models.base.Model.Admin):
        search_fields = ["url"]
        exclude = ("extra_options",)

        @classmethod
        def _mbd_get_or_create(cls, request, url):
            if url:
                Archive.mbd_get_or_create(request, url=url)

        @classmethod
        def mbd_meta_add_local(cls, request):
            """
            Local scan for archives.

            This currently scans the local sources list and tries to detect a local apt-cacher-ng.
            """
            try:
                import aptsources.sourceslist
                for src in (src for src in aptsources.sourceslist.SourcesList() if not src.invalid and not src.disabled):
                    # These URLs come from the user. 'normalize' the uri first to have exactly one trailing slash.
                    cls._mbd_get_or_create(request, src.uri.rstrip("/") + "/")

            except BaseException as e:
                mini_buildd.log_exception(LOG, "Failed to scan local sources.lists for default mirrors ('python-apt' not installed?)", e)

            url = mini_buildd.net.detect_apt_cacher_ng(url=f"http://{socket.getfqdn()}:3142")
            if url:
                MsgLog(LOG, request).info(f"Local apt-cacher-ng detected: {url}")
                for path in ["debian", "ubuntu", "debian-security", "debian-archive/debian", "debian-archive/debian-security", "debian-archive/debian-backports"]:
                    cls._mbd_get_or_create(request, f"{url}/{path}/")

        @classmethod
        def mbd_meta_add_debian(cls, request):
            """Add internet Debian archive sources."""
            for url in ["http://ftp.debian.org/debian/",                 # Debian (release, updates, proposed-updates and backports)
                        "http://deb.debian.org/debian/",                 # alternate: CDN

                        "http://security.debian.org/debian-security/",   # Debian Security (release/updates)
                        "http://deb.debian.org/debian-security/",        # alternate: CDN

                        "http://archive.debian.org/debian/",             # Archived Debian Releases
                        "http://archive.debian.org/debian-security/",    # Archived Debian Security
                        "http://archive.debian.org/debian-backports/",   # Archived Debian Backports
                        ]:
                cls._mbd_get_or_create(request, url)
            MsgLog(LOG, request).info("Consider adding archives with your local or closest mirrors; check 'netselect-apt'.")

        @classmethod
        def mbd_meta_add_ubuntu(cls, request):
            """Add internet Ubuntu archive sources."""
            for url in ["http://archive.ubuntu.com/ubuntu/",              # Ubuntu releases
                        "http://security.ubuntu.com/ubuntu/",             # Ubuntu Security
                        "http://old-releases.ubuntu.com/ubuntu/",         # Older Ubuntu release
                        ]:
                cls._mbd_get_or_create(request, url)
            MsgLog(LOG, request).info("Consider replacing these archives with you closest mirror(s); check netselect-apt.")

    def __str__(self):
        return f"{self.url} (ping {round(self.ping, 2)} ms)"

    # Note: pylint false-positive: https://github.com/PyCQA/pylint/issues/1553
    def clean(self, *args, **kwargs):  # pylint: disable=arguments-differ
        super().clean(*args, **kwargs)
        if self.url[-1] != "/" or self.url[-2] == "/":  # pylint: disable=unsubscriptable-object
            raise django.core.exceptions.ValidationError({"url": "The URL must have exactly one trailing slash (like 'http://example.org/path/')."})

    def mbd_get_matching_release(self, request, source, gnupg):
        def download(file_, url):
            file_.write(urllib.request.urlopen(url, timeout=self.URLOPEN_TIMEOUT).read())
            file_.flush()
            file_.seek(0)

        url_base = f"{self.url}/dists/{source.codename}/"
        with tempfile.NamedTemporaryFile() as release_file, tempfile.NamedTemporaryFile() as signature_file:
            signature_url = None
            try:
                release_url = url_base + "InRelease"
                download(release_file, release_url)
            except urllib.error.HTTPError:
                try:
                    release_url = url_base + "Release"
                    download(release_file, release_url)
                    signature_url = release_url + ".gpg"
                    download(signature_file, signature_url)
                except urllib.error.HTTPError:
                    raise mini_buildd.HTTPUnavailable(f"{url_base}: Neither 'InRelease' nor 'Release[.gpg]' found")

            MsgLog(LOG, request).debug(f"Checking: {release_url} for {source}")
            release = debian.deb822.Release(release_file)

            # Check release file fields
            if not source.mbd_is_matching_release(release):
                raise Exception(f"{release_url} does not match {source}')")

            # Pre-Check 'Valid-Until'
            #
            # Some Release files contain an expire date via the
            # 'Valid-Until' tag. If such an expired archive is used,
            # builds will fail. Furthermore, it could be only the
            # selected archive not updating, while the source may be
            # perfectly fine from another archive.
            #
            # This pre-check avoids such a situation, or at least it
            # can be fixed by re-checking the source.
            try:
                valid_until = release["Valid-Until"]
                if dateutil.parser.parse(valid_until) < mini_buildd.misc.datetime_stamp():
                    if source.mbd_get_extra_option("X-Check-Valid-Until", "yes").lower() in ("no", "false", "0"):
                        MsgLog(LOG, request).warning(f"{release_url} expired, but source marked to ignore valid-until (Valid-Until='{valid_until}').")
                    else:
                        MsgLog(LOG, request).warning(f"{release_url} expired, maybe the archive has problems? (Valid-Until='{valid_until}').")
                        raise Exception(f"{release_url} expired")
            except KeyError:
                pass  # We can assume Release file has no "Valid-Until", and be quiet
            except BaseException as e:
                mini_buildd.log_exception(LOG, f"Ignoring error checking 'Valid-Until' on {release_url}", e)

            # Check signature
            sigs = gnupg.verify2(release_file, signature_file if signature_url is not None else None, accept_expired=True)

            # User success info. This also prints the list of *all* keys the release file was signed with (possibly hinting to clean up setup)
            MsgLog(LOG, request).info(f"'{source}' found via '{release_url}' ({gnupg.gpgme_sigs2info(sigs)})")

            # Ok, this is for this source
            return release

    def mbd_ping(self):
        """Ping and update the ping value."""
        try:
            t0 = django.utils.timezone.now()
            # Append dists to URL for ping check: Archive may be
            # just fine, but not allow to access to base URL
            # (like ourselves ;). Any archive _must_ have dists/ anyway.
            try:
                urllib.request.urlopen(f"{self.url}/dists/", timeout=self.URLOPEN_TIMEOUT)
            except urllib.error.HTTPError as e:
                # Allow HTTP 4xx client errors through; these might be valid use cases like:
                # 404 Usage Information: apt-cacher-ng
                if not 400 <= e.code <= 499:
                    raise

            delta = django.utils.timezone.now() - t0
            self.ping = delta.total_seconds() * (10 ** 3)
            self.save()
        except Exception as e:
            self.ping = -1.0
            self.save()
            raise Exception(f"{self}: Does not ping: {e}")

    def mbd_get_reverse_dependencies(self):
        """Return all sources (and their deps) that use us."""
        result = list(self.source_set.all())
        for s in self.source_set.all():
            result += s.mbd_get_reverse_dependencies()
        return result


class Architecture(mini_buildd.models.base.Model):
    name = django.db.models.CharField(primary_key=True, max_length=50)

    def __str__(self):
        return f"{self.name}"

    @classmethod
    def mbd_host_architecture(cls):
        return mini_buildd.call.Call(["dpkg", "--print-architecture"]).check().stdout.strip()

    @classmethod
    def mbd_supported_architectures(cls, arch=None):
        """Get all supported architectures (some archs also natively support other archs)."""
        arch = arch or cls.mbd_host_architecture()
        arch_map = {"amd64": ["i386"]}
        return [arch] + arch_map.get(arch, [])


@functools.total_ordering
class Component(mini_buildd.models.base.Model):
    """
    A Debian component (like 'main', 'contrib', 'non-free').

    Sorting: 'main' should be first, the others in no special
    order. You may just use standard python sorted() on an iterable
    of components.
    """

    name = django.db.models.CharField(primary_key=True, max_length=50)

    def __str__(self):
        return str(self.name)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.name == other.name
        return NotImplemented

    def __hash__(self):  # pylint: disable=useless-super-delegation
        return super().__hash__()

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            return self.name == "main"
        return NotImplemented


@functools.total_ordering
class Source(mini_buildd.models.base.StatusModel):
    # Identity
    origin = django.db.models.CharField(max_length=60, default="Debian",
                                        help_text="The exact string of the 'Origin' field of the resp. Release file.")
    codename = django.db.models.CharField(max_length=60, default="sid",
                                          help_text=("<p>The <b>name of the directory below <code>dists/</code></b> in archives "
                                                     "this source refers to. This is also the 3rd part of an apt line."
                                                     "</p><p>"
                                                     "With no extra options given, this source will be identified "
                                                     "comparing <code>Origin</code> and <code>Codename</code> with the values of the "
                                                     "<code>Release</code> file found."
                                                     "</p><p>"
                                                     "Some sources need some more care via <em>Extra options</em> below:</p>"
                                                     ""
                                                     "<b>Origin, Codename, Suite, Archive, Version, Label</b>:"
                                                     ""
                                                     "<p>If needed, you may use these fields (same as in a Release file) to "
                                                     "further specify this source. These are also later used to pin the "
                                                     "source via apt.</p>"
                                                     ""
                                                     "<p>For some sources, <code>Codename</code> (as we use it above) does "
                                                     "not match resp. value as given in the Release file. When "
                                                     "<code>Codename</code> is overridden in this manner, be sure to also "
                                                     "add one further flag to identify the source -- else apt pinning later "
                                                     "would likely not be unambiguous. Real world examples that need this "
                                                     "extra handling are <em>Debian Security</em> (<code>buster</code> or "
                                                     "older), <em>Ubuntu Security and Backports</em>:</p>"
                                                     ""
                                                     "<pre>"
                                                     "Codename: bionic\n"
                                                     "Suite: bionic-backports\n"
                                                     "</pre>"
                                                     ""
                                                     "<b>X-Check-Valid-Until</b>:"
                                                     ""
                                                     "<p>Some sources have a <code>Valid-Until</code> field that is no longer"
                                                     "updated. If you really still want to use it anyway, use:</p>"
                                                     ""
                                                     "<pre>"
                                                     "X-Check-Valid-Until: no"
                                                     "</pre>"
                                                     ""
                                                     "<p>This will, 1st, ignore mini-buildd's own 'Valid-Until check' and "
                                                     "2nd, create apt lines for this source with the "
                                                     "<code>[check-valid-until=no]</code> option. I.e., at least from "
                                                     "<code>stretch</code> onwards, the check is disabled <em>per "
                                                     "source</em>. For <code>jessie</code> or worse (where this apt option "
                                                     "does not work), a global workaround via schroot is still in place.</p>"
                                                     ""
                                                     "<b>X-Remove-From-Component</b>:"
                                                     ""
                                                     "<p>Some (actually, we only know of <em>Debian Security</em>) sources "
                                                     "have weird <code>Components</code> that need to be fixed to work with "
                                                     "mini-buildd. For example, <em>Debian Security</em> (buster or older) "
                                                     "needs:</p>"
                                                     ""
                                                     "<pre>"
                                                     "Codename: stretch\n"
                                                     "Label: Debian-Security\n"
                                                     "X-Remove-From-Component: updates/\n"
                                                     "</pre>"
                                                     ""
                                                     "<p>Fwiw, in <em>Debian Security</em> for <code>bullseye</code> or "
                                                     "younger, Codename is now unambiguous (like "
                                                     "<code>bullseye-security</code>), but the component oddity remains; so "
                                                     "these cases still need:</p>"
                                                     ""
                                                     "<pre>"
                                                     "X-Remove-From-Component: updates/"
                                                     "</pre>"))

    # Apt Secure
    apt_keys = django.db.models.ManyToManyField(mini_buildd.models.gnupg.AptKey, blank=True,
                                                help_text=("At least one key this source's 'Release' file is signed with (Hint: If you leave this empty and run 'check' on it, you will get a message listing all signer keys)."))

    # Extra
    description = django.db.models.CharField(max_length=100, editable=False, blank=True, default="")
    codeversion = django.db.models.CharField(max_length=50, editable=False, blank=True, default="")
    codeversion_override = django.db.models.CharField(
        max_length=50, blank=True, default="",
        help_text=("Save this as empty string to have the codeversion re-guessed on check, or\n"
                   "put your own override value here if the guessed string is broken. The\n"
                   "codeversion is only used for base sources.\n"))
    archives = django.db.models.ManyToManyField(Archive, blank=True)
    components = django.db.models.ManyToManyField(Component, blank=True)
    architectures = django.db.models.ManyToManyField(Architecture, blank=True)

    class Meta(mini_buildd.models.base.StatusModel.Meta):
        unique_together = ("origin", "codename")
        ordering = ["origin", "-codeversion", "codename"]

    class Admin(mini_buildd.models.base.StatusModel.Admin):
        list_display = mini_buildd.models.base.StatusModel.Admin.list_display + ["origin", "codeversion", "codename"]
        search_fields = ["origin", "codeversion", "codename"]
        ordering = ["origin", "-codeversion", "codename"]

        readonly_fields = ["codeversion", "archives", "components", "architectures", "description"]
        fieldsets = (
            ("Identity", {"fields": ("origin", "codename", "extra_options", "apt_keys")}),
            ("Extra", {"classes": ("collapse",), "fields": ("description", "codeversion", "codeversion_override", "archives", "components", "architectures")}),)
        filter_horizontal = ("apt_keys",)

        def get_readonly_fields(self, _request, obj=None):
            """Forbid to change identity on existing source (usually a bad idea; repos/chroots that refer to us may break)."""
            fields = copy.copy(self.readonly_fields)
            if obj:
                fields.append("origin")
                fields.append("codename")
            return fields

        @classmethod
        def _mbd_get_or_create(cls, request, origin, codename, keys, extra_options=""):
            try:
                obj, created = Source.mbd_get_or_create(request, origin=origin, codename=codename, extra_options=extra_options)
                if created:
                    for long_key_id in keys:
                        matching_keys = mini_buildd.models.gnupg.AptKey.mbd_filter_key(long_key_id)
                        if matching_keys:
                            apt_key = matching_keys[0]
                        else:
                            apt_key, _created = mini_buildd.models.gnupg.AptKey.mbd_get_or_create(request, key_id=long_key_id)
                        obj.apt_keys.add(apt_key)
                    obj.save()
            except BaseException as e:
                LOG.debug(f"Can't add {codename} (most likely a non-default instance already exists): {e}")

        @classmethod
        def mbd_meta_add_debian(cls, request):
            """
            Add well-known Debian sources.

            To display the key ids via apt-key in the format as used here::

              apt-key adv --list-public-keys --keyid-format=long

            """
            keys = {
                "archive_buster": "DC30D7C23CBBABEE",            # Debian Archive Automatic Signing Key (10/buster) <ftpmaster@debian.org>
                "release_buster": "DCC9EFBF77E11517",            # Debian Stable Release Key (10/buster) <debian-release@lists.debian.org>
                "archive_stretch": "E0B11894F66AEC98",           # Debian Archive Automatic Signing Key (9/stretch) <ftpmaster@debian.org>  (subkey 04EE7237B7D453EC)
                "release_stretch": "EF0F382A1A7B6500",           # Debian Stable Release Key (9/stretch) <debian-release@lists.debian.org>
                "archive_jessie": "7638D0442B90D010",            # Debian Archive Automatic Signing Key (8/jessie) <ftpmaster@debian.org>
                "release_jessie": "CBF8D6FD518E17E1",            # Jessie Stable Release Key <debian-release@lists.debian.org>
                "security_archive_jessie": "9D6D8F6BC857C906",   # Debian Security Archive Automatic Signing Key (8/jessie) <ftpmaster@debian.org>
                "security_archive_stretch": "EDA0D2388AE22BA9",  # Debian Security Archive Automatic Signing Key (9/stretch) <ftpmaster@debian.org>
                "security_archive_buster": "4DFAB270CAA96DFA",   # Debian Security Archive Automatic Signing Key (10/buster) <ftpmaster@debian.org>
                "archive_wheezy": "8B48AD6246925553",            # Debian Archive Automatic Signing Key (7.0/wheezy) <ftpmaster@debian.org>
                "release_wheezy": "6FB2A1C265FFB764",            # Wheezy Stable Release Key <debian-release@lists.debian.org>
            }

            cls._mbd_get_or_create(request, "Debian", "wheezy",
                                   [keys["archive_wheezy"], keys["release_wheezy"], keys["archive_jessie"]])
            cls._mbd_get_or_create(request, "Debian", "wheezy/updates",
                                   [keys["archive_wheezy"], keys["security_archive_jessie"]],
                                   extra_options="Codename: wheezy\nLabel: Debian-Security\nX-Remove-From-Component: updates/\nX-Check-Valid-Until: no")
            cls._mbd_get_or_create(request, "Debian Backports", "wheezy-backports",
                                   [keys["archive_wheezy"], keys["release_wheezy"], keys["archive_jessie"]])
            cls._mbd_get_or_create(request, "Debian Backports", "wheezy-backports-sloppy",
                                   [keys["archive_wheezy"], keys["archive_jessie"]])

            cls._mbd_get_or_create(request, "Debian", "jessie",
                                   [keys["archive_wheezy"], keys["release_jessie"], keys["archive_jessie"]])
            cls._mbd_get_or_create(request, "Debian", "jessie/updates",
                                   [keys["security_archive_jessie"], keys["security_archive_stretch"]],
                                   extra_options="Codename: jessie\nLabel: Debian-Security\nX-Remove-From-Component: updates/")
            cls._mbd_get_or_create(request, "Debian Backports", "jessie-backports",
                                   [keys["archive_wheezy"], keys["archive_jessie"]],
                                   extra_options="X-Check-Valid-Until: no")
            cls._mbd_get_or_create(request, "Debian Backports", "jessie-backports-sloppy",
                                   [keys["archive_wheezy"], keys["archive_jessie"]],
                                   extra_options="X-Check-Valid-Until: no")

            cls._mbd_get_or_create(request, "Debian", "stretch",
                                   [keys["archive_jessie"], keys["archive_stretch"], keys["release_stretch"]])
            cls._mbd_get_or_create(request, "Debian", "stretch/updates",
                                   [keys["security_archive_buster"], keys["security_archive_stretch"]],
                                   extra_options="Codename: stretch\nLabel: Debian-Security\nX-Remove-From-Component: updates/")
            cls._mbd_get_or_create(request, "Debian Backports", "stretch-backports",
                                   [keys["archive_buster"], keys["archive_stretch"]])
            cls._mbd_get_or_create(request, "Debian Backports", "stretch-backports-sloppy",
                                   [keys["archive_buster"], keys["archive_stretch"]])

            cls._mbd_get_or_create(request, "Debian", "buster",
                                   [keys["release_stretch"], keys["release_buster"], keys["archive_stretch"], keys["archive_buster"]])
            cls._mbd_get_or_create(request, "Debian", "buster/updates",
                                   [keys["security_archive_stretch"], keys["security_archive_buster"]],
                                   extra_options="Codename: buster\nLabel: Debian-Security\nX-Remove-From-Component: updates/")
            cls._mbd_get_or_create(request, "Debian Backports", "buster-backports",
                                   [keys["archive_stretch"], keys["archive_buster"]])

            cls._mbd_get_or_create(request, "Debian", "bullseye",
                                   [keys["archive_stretch"], keys["archive_buster"]])
            cls._mbd_get_or_create(request, "Debian", "bullseye-security",
                                   [keys["security_archive_stretch"], keys["security_archive_buster"]],
                                   extra_options="X-Remove-From-Component: updates/")

            cls._mbd_get_or_create(request, "Debian", "sid",
                                   [keys["archive_stretch"], keys["archive_buster"]])
            cls._mbd_get_or_create(request, "Debian", "experimental",
                                   [keys["archive_stretch"], keys["archive_buster"]])

        @classmethod
        def mbd_meta_add_ubuntu(cls, request):
            """Add well-known Ubuntu sources. Update hint: Keep latest two releases plus a couple of LTS releases."""
            keys = {
                "archive_current": "40976EAF437D05B5",    # Ubuntu Archive Automatic Signing Key <ftpmaster@ubuntu.com>
                "archive_2012": "3B4FE6ACC0B21F32",       # Ubuntu Archive Automatic Signing Key (2012) <ftpmaster@ubuntu.com>
                "archive_2018": "871920D1991BC93C",       # Ubuntu Archive Automatic Signing Key (2018) <ftpmaster@ubuntu.com>
            }

            # xenial: 16.04 (LTS until 2021)
            cls._mbd_get_or_create(request, "Ubuntu", "xenial",
                                   [keys["archive_current"], keys["archive_2012"]])
            cls._mbd_get_or_create(request, "Ubuntu", "xenial-security",
                                   [keys["archive_current"], keys["archive_2012"]],
                                   "Codename: xenial\nSuite: xenial-security")
            cls._mbd_get_or_create(request, "Ubuntu", "xenial-backports",
                                   [keys["archive_current"], keys["archive_2012"]],
                                   "Codename: xenial\nSuite: xenial-backports")

            # bionic: 18.04 (LTS until 2023)
            cls._mbd_get_or_create(request, "Ubuntu", "bionic",
                                   [keys["archive_current"], keys["archive_2012"]])
            cls._mbd_get_or_create(request, "Ubuntu", "bionic-security",
                                   [keys["archive_current"], keys["archive_2012"]],
                                   "Codename: bionic\nSuite: bionic-security")
            cls._mbd_get_or_create(request, "Ubuntu", "bionic-backports",
                                   [keys["archive_current"], keys["archive_2012"]],
                                   "Codename: bionic\nSuite: bionic-backports")

            # eoan: 19.10
            cls._mbd_get_or_create(request, "Ubuntu", "eoan",
                                   [keys["archive_2012"], keys["archive_2018"]])
            cls._mbd_get_or_create(request, "Ubuntu", "eoan-security",
                                   [keys["archive_2012"], keys["archive_2018"]],
                                   "Codename: eoan\nSuite: eoan-security")
            cls._mbd_get_or_create(request, "Ubuntu", "eoan-backports",
                                   [keys["archive_2012"], keys["archive_2018"]],
                                   "Codename: eoan\nSuite: eoan-backports")

            # focal: 20.04 (LTS until 2025)
            cls._mbd_get_or_create(request, "Ubuntu", "focal",
                                   [keys["archive_2012"], keys["archive_2018"]])
            cls._mbd_get_or_create(request, "Ubuntu", "focal-security",
                                   [keys["archive_2012"], keys["archive_2018"]],
                                   "Codename: focal\nSuite: focal-security")
            cls._mbd_get_or_create(request, "Ubuntu", "focal-backports",
                                   [keys["archive_2012"], keys["archive_2018"]],
                                   "Codename: focal\nSuite: focal-backports")

            # groovy: 20.10
            cls._mbd_get_or_create(request, "Ubuntu", "groovy",
                                   [keys["archive_2012"], keys["archive_2018"]])
            cls._mbd_get_or_create(request, "Ubuntu", "groovy-security",
                                   [keys["archive_2012"], keys["archive_2018"]],
                                   "Codename: groovy\nSuite: groovy-security")
            cls._mbd_get_or_create(request, "Ubuntu", "groovy-backports",
                                   [keys["archive_2012"], keys["archive_2018"]],
                                   "Codename: groovy\nSuite: groovy-backports")

        @classmethod
        def mbd_filter_active_base_sources(cls):
            """Filter active base sources; needed in chroot and distribution wizards."""
            return Source.objects.filter(status__gte=Source.STATUS_ACTIVE,
                                         origin__in=["Debian", "Ubuntu"],
                                         codename__regex=r"^[a-z]+$").exclude(codename="experimental")

    def __str__(self):
        return f"{self.origin} {self.codeversion} ({self.codename})"

    def mbd_release_file_values(self):
        """Compute a dict of values a matching release file must have."""
        values = {k: v for k, v in self.mbd_get_extra_options().items() if not k.startswith("X-")}  # Keep "X-<header>" for special purposes. All other keys are like in a Release file.

        # Set Origin and Codename (may be overwritten) from fields
        values["Origin"] = self.origin
        if not values.get("Codename"):
            values["Codename"] = self.codename

        return values

    def mbd_is_matching_release(self, release):
        """Check that this release file matches us."""
        for key, value in list(self.mbd_release_file_values().items()):
            # Check identity: origin, codename
            if value != release[key]:
                return False
        return True

    def mbd_get_archive(self):
        """Get fastest archive."""
        oa_list = self.archives.all().filter(ping__gte=0.0).order_by("ping")
        if oa_list:
            return oa_list[0]
        raise mini_buildd.HTTPBadRequest(f"{self}: No archive found. Please add appropriate archive and/or check network setup.")

    def mbd_get_apt_line(self, limit_components=None):
        components = [c for c in self.components.all() if limit_components is None or c in limit_components]
        return mini_buildd.files.AptLine(self.mbd_get_archive().url,
                                         "",
                                         self.codename,
                                         components,
                                         options=f"check-valid-until={self.mbd_get_extra_option('X-Check-Valid-Until', 'yes')}",
                                         comment=f"{self}: {self.mbd_get_apt_pin()}")

    def mbd_get_apt_pin(self):
        """Apt 'pin line' (for use in a apt 'preference' file)."""
        # See man apt_preferences for the field/pin mapping
        supported_fields = {"Origin": "o", "Codename": "n", "Suite": "a", "Archive": "a", "Version": "v", "Label": "l"}
        pins = []
        for key, value in list(self.mbd_release_file_values().items()):
            k = supported_fields.get(key)
            if k:
                pins.append(f"{k}={value}")
        return "release " + ", ".join(pins)

    def mbd_prepare(self, _request):
        self.clean()

    def mbd_sync(self, request):
        self._mbd_remove_and_prepare(request)

    def mbd_remove(self, _request):
        self.archives.set([])
        self.components.set([])
        self.architectures.set([])
        self.description = ""

    def mbd_check(self, request):
        """Rescan all archives, and check that there is at least one working."""
        self.archives.set([])
        with contextlib.closing(mini_buildd.gnupg.TmpGnuPG(tmpdir_options={"prefix": "gnupg-sourcecheck-"})) as gpg:
            for k in self.apt_keys.all():
                gpg.add_pub_key(k.key)

            for archive in Archive.objects.all():
                try:
                    # Check and update the ping value
                    archive.mbd_ping()

                    # Get release if this archive serves us, else exception
                    release = archive.mbd_get_matching_release(request, self, gpg)

                    # Implicitly save ping value for this archive
                    self.archives.add(archive)
                    self.description = release["Description"]

                    # Set codeversion
                    self.codeversion = ""
                    if self.codeversion_override:
                        self.codeversion = self.codeversion_override
                    else:
                        self.codeversion = mini_buildd.distribution.guess_codeversion(release)
                        self.codeversion_override = self.codeversion

                    # Set architectures and components (may be auto-added)
                    if release.get("Architectures"):
                        for a in release["Architectures"].split(" "):
                            new_arch, _created = Architecture.mbd_get_or_create(request, name=a)
                            self.architectures.add(new_arch)
                    if release.get("Components"):
                        for c in release["Components"].split(" "):
                            new_component, _created = Component.mbd_get_or_create(request, name=c.replace(self.mbd_get_extra_option("X-Remove-From-Component", ""), ""))
                            self.components.add(new_component)
                except mini_buildd.HTTPUnavailable as e:
                    MsgLog(LOG, request).debug(f"'{self}' not hosted by '{archive}': {e}")
                except mini_buildd.HTTPUnauthorized as e:
                    MsgLog(LOG, request).warning(f"'{self}' hosted by '{archive}', but unauthorized: {e}")
                except BaseException as e:
                    MsgLog(LOG, request).error(f"'{self}' error with '{archive}': {mini_buildd.e2http(e)}")
                    mini_buildd.log_exception(LOG, f"'{self}' error with '{archive}'", e)

        # Check that at least one archive can be found
        fastest_archive = self.mbd_get_archive()
        MsgLog(LOG, request).info(f"{self}: Fastest archive: {fastest_archive}")

    def mbd_get_dependencies(self):
        return self.apt_keys.all()

    def mbd_get_reverse_dependencies(self):
        """Return all chroots and repositories that use us."""
        result = list(self.chroot_set.all())
        for d in self.distribution_set.all():
            result += d.mbd_get_reverse_dependencies()
        return result

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.origin == other.origin and self.codename == other.codename
        return NotImplemented

    def __hash__(self):  # pylint: disable=useless-super-delegation
        return super().__hash__()

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            if self.origin == other.origin:
                try:
                    return int(self.codeversion) < int(other.codeversion)
                except BaseException:
                    return self.codeversion < other.codeversion
            return self.origin < other.origin
        return NotImplemented

    # Note: pylint false-positive: https://github.com/PyCQA/pylint/issues/1553
    def clean(self, *args, **kwargs):  # pylint: disable=arguments-differ
        super().clean(*args, **kwargs)
        if self.mbd_get_extra_option("Origin"):
            raise django.core.exceptions.ValidationError({"extra_options": "You may not override 'Origin', just use the origin field."})


class PrioritySource(mini_buildd.models.base.Model):
    source = django.db.models.ForeignKey(Source,
                                         on_delete=django.db.models.CASCADE)
    priority = django.db.models.IntegerField(default=1,
                                             help_text="A apt pin priority value (see 'man apt_preferences')."
                                             "Examples: 1=not automatic, 1001=downgrade'")

    class Meta(mini_buildd.models.base.Model.Meta):
        unique_together = ('source', 'priority')

    class Admin(mini_buildd.models.base.Model.Admin):
        exclude = ("extra_options",)

        @classmethod
        def mbd_meta_add_extras(cls, request):
            """Smartly add priority sources from 'non-base' sources: Security gets prio=500, others prio=1 (opt-in)."""
            def _create(source, prio):
                prio_source = PrioritySource.objects.filter(source=source).first()
                if not prio_source:
                    PrioritySource.mbd_get_or_create(request, source=source, priority=prio)

            for source in Source.objects.filter(origin="Debian", codename__regex="^[a-z]+-security$"):  # Security: Debian uses this scheme for security since "bullseye"
                _create(source, 500)
            for source in Source.objects.filter(origin="Debian", codename__regex="^[a-z]+/updates$"):   # Security: Before "bullseye", Debian used strange CODENAME/updates scheme
                _create(source, 500)
            for source in Source.objects.filter(origin="Ubuntu", codename__regex="^[a-z]+-security$"):  # Security: Ubuntu
                _create(source, 500)
            for source in Source.objects.exclude(codename__regex=r"^[a-z]+$"):                          # All other sources not looking like a base codename should get prio=1 (opt-in)
                _create(source, 1)
            for source in Source.objects.filter(origin="Debian", codename="experimental"):              # Debian "experimental" special case
                _create(source, 1)

    def __str__(self):
        return f"{self.source} with prio={self.priority}"

    def mbd_get_apt_preferences(self):
        return f"Package: *\nPin: {self.source.mbd_get_apt_pin()}\nPin-Priority: {self.priority}\n"
