import os
import queue
import logging

import mini_buildd.misc
import mini_buildd.changes
import mini_buildd.threads


LOG = logging.getLogger(__name__)


class Package():
    def __init__(self, event):
        self.upload = mini_buildd.changes.Upload(event)
        self.running = mini_buildd.misc.StopWatch()
        self.status = None
        self.extra = {}

        self.requests, self.success, self.failed = {}, {}, {}

        # Get/check repository, distribution and suite for changes
        self.repository, self.distribution, self.suite, _rollback = mini_buildd.models.repository.parse_distribution(self.upload.distribution.get(), check_uploadable=True)

        # Authenticate
        if self.repository.allow_unauthenticated_uploads:
            LOG.warning(f"Unauthenticated uploads allowed. Using '{self.upload.file_name}' unchecked")
        else:
            mini_buildd.Daemon().keyrings.get_uploaders()[self.repository.identity].verify(self.upload.file_path)

        # Repository package prechecks
        self.repository.mbd_package_precheck(self.distribution, self.suite, self.upload["source"], self.upload["version"])

    def __str__(self):
        return self.upload.key

    def upload_buildrequests(self):
        # Generate build requests
        self.requests = self.upload.gen_buildrequests(self.repository, self.distribution, self.suite)

        # Upload buildrequests
        for _key, buildrequest in list(self.requests.items()):
            buildrequest.upload_buildrequest(mini_buildd.http_endpoint())

    def add_buildresult(self, buildresult):
        mini_buildd.Daemon().keyrings.get_remotes().verify(buildresult.file_path)

        arch = buildresult["Architecture"]

        # Retval and status must be the same with sbuild in mode user, so status is not really needed
        # status may also be none in case some non-sbuild build error occurred
        retval = int(buildresult.cget("Sbuild-Retval"))
        status = buildresult.cget("Sbuild-Status")
        lintian = buildresult.cget("Sbuild-Lintian")
        piuparts = buildresult.cget("Sbuild-Piuparts")
        autopkgtest = buildresult.cget("Sbuild-Autopkgtest")

        LOG.info(f"{self.upload.key}: Got build result for '{arch}': {retval}={status}, lintian={lintian}")

        def check(arch, status, checker, mode):
            """
            Check if arch is allowed to succeed based on check results.

            Succeeds if
            - the actual check run has no errors or warnings
            - the distribution is configured to ignore check results
            - upload says to ignore check results
            - suite is experimental (we deliberately ignore check in that case)
            - warnings are present, but check mode is not warnfail
            """
            check_mode = self.upload.check_mode(checker, arch, mode)
            return check_mode.check(status, ignore=self.suite.experimental)

        def check_all():
            return \
                check(arch, lintian, "lintian", self.distribution.lintian_mode) and \
                check(arch, piuparts, "piuparts", self.distribution.piuparts_mode) and \
                check(arch, autopkgtest, "autopkgtest", int(self.distribution.mbd_get_extra_option("Autopkgtest-Mode", "0")))

        if retval == 0 and (status == "skipped" or check_all()):
            self.success[arch] = buildresult
        else:
            self.failed[arch] = buildresult

        missing = len(self.requests) - len(self.success) - len(self.failed)
        if missing <= 0:
            try:
                self.repository.mbd_package_install(self.distribution, self.suite, self.upload, self.success)
                self.status = mini_buildd.events.Type.INSTALLED
                self.auto_ports()
            except BaseException as e:
                self.status = mini_buildd.events.Type.FAILED
                self.extra["error"] = mini_buildd.e2http(e).rfc7807.to_json()
            finally:
                self.extra["log"] = self.upload.get_log_path().path
                self.extra["buildresults"] = {buildresult.get("architecture"): buildresult.to_json() for buildresult in list(self.success.values()) + list(self.failed.values())}
                event = mini_buildd.events.Event.from_changes(self.status, self.upload, extra=self.extra)
                self.move_to_pkglog()
                self.running.close()
                mini_buildd.Daemon().events.log(event)

    def auto_ports(self):
        """Try to serve all automatic ports. This will never fail -- failures will be event-logged only."""
        for to_dist_str in self.upload.options.get("auto-ports", default=[]):
            self.extra.setdefault("auto_ports", {})
            try:
                self.extra["auto_ports"][to_dist_str] = (True, mini_buildd.Daemon().port(self.upload["Source"],
                                                                                         self.upload.distribution.get(),
                                                                                         to_dist_str,
                                                                                         self.upload["Version"]))
            except BaseException as e:
                self.extra["auto_ports"][to_dist_str] = (False, f"{e}")

    def purge_spool(self):
        # Purge complete package spool dir (if precheck failed, spool dir will not be present, so we need to ignore errors here)
        mini_buildd.misc.skip_if_keep_in_debug(mini_buildd.misc.rmdirs, self.upload.get_spool_dir())

    def move_to_pkglog(self):
        self.upload.move_to_pkglog(self.status == mini_buildd.events.Type.INSTALLED, rejected=self.status == mini_buildd.events.Type.REJECTED)

        for _arch, c in list(self.success.items()) + list(self.failed.items()) + list(self.requests.items()):
            c.move_to_pkglog(self.status == mini_buildd.events.Type.INSTALLED, rejected=self.status == mini_buildd.events.Type.REJECTED)

        self.purge_spool()


class Packager(mini_buildd.threads.EventThread):
    def __init__(self):
        self.queue = queue.Queue()
        super().__init__(events=self.queue, name="packager")
        self.packaging = {}

    def run_event(self, event):
        try:
            with self.lock:
                if mini_buildd.changes.Buildresult.match(event):
                    buildresult = mini_buildd.changes.Buildresult(event)
                    if buildresult.key not in self.packaging:
                        raise mini_buildd.HTTPBadRequest(f"Stray buildresult (not packaging here): {buildresult.bkey}")

                    package = self.packaging[buildresult.key]
                    package.add_buildresult(buildresult)
                    if package.running.stop:
                        del self.packaging[package.upload.key]

                else:
                    package = mini_buildd.packager.Package(event)
                    if package.upload.key in self.packaging:
                        raise mini_buildd.HTTPInternal(f"Stray upload (already packaging): {package.upload.key}")  # Incoming handler should always prevent this in the 1st place
                    package.upload_buildrequests()
                    self.packaging[package.upload.key] = package

        except BaseException as e:
            # Try hard to clean up bogus upload and log REJECTED
            changes = None
            try:
                changes = mini_buildd.changes.Changes(event)
                changes.remove()
            except BaseException:
                mini_buildd.misc.attempt(os.remove, event)
            finally:
                mini_buildd.Daemon().events.logchanges(mini_buildd.events.Type.REJECTED, {} if changes is None else changes, extra={"error": mini_buildd.e2http(e).rfc7807.to_json()})

    def terminate(self, key):
        with self.lock:
            package = self.packaging.get(key)
            if package:
                del self.packaging[key]
                package.purge_spool()
            else:
                raise mini_buildd.HTTPBadRequest(f"No such package packaging: {key}")
