import mini_buildd.files


class Blocks(mini_buildd.files.Dir):
    def __init__(self, type_):
        super().__init__()
        self.type = type_

    def extra_option(self, top=False):
        return f"Sbuild-{self.type.capitalize()}-Blocks" + ("-Top" if top else "")

    def extra_options(self):
        return [self.extra_option(top=True), self.extra_option()]

    def usage(self):
        ml = max([len(name) for name in self.extra_options()])  # "max length" helper
        header = (f"{self.extra_option(top=True):<{ml}}: Blocks before the automated {self.type}.\n"
                  f"{self.extra_option(top=False):<{ml}}: Blocks after the automated {self.type}.\n\n"
                  "Available blocks:\n\n")
        ml = max([len(name) for name in self.keys()])
        return header + "\n".join([f"{name:<{ml}}: {block.description}" for name, block in self.items()])

    def validate(self, extra_option):
        """Validate extra option value from user space (string, space separated). Return extra option value as list."""
        eol = extra_option.split()  # extra option list
        unknown = set(eol) - set(self.keys())
        if unknown:
            raise mini_buildd.HTTPBadRequest(f"Unknown sbuild {self.type} block(s): {' '.join(unknown)} (check 'Extra Options' configurations in Distributions)")
        return eol

    def validate_all(self, get):
        for extra_option in self.extra_options():
            self.validate(get(extra_option, ""))


class ConfigBlocks(Blocks):
    def __init__(self):
        super().__init__("config")

    def default(self):
        return f"{self.extra_option(top=False)}: ccache\n"

    def configure(self, file_, get, top):
        """Add configured blocks to provided file (space separated config option is auto-magically received via provided getter)."""
        for block in self.validate(get(self.extra_option(top=top), "")):
            file_.add(self[block].get(snippet=True))


class SetupBlocks(Blocks):
    def __init__(self):
        super().__init__("setup")

    def default(self):
        return (f"{self.extra_option(top=True)}: disable-check-valid-until\n"
                f"{self.extra_option(top=False)}: ccache eatmydata\n")

    def configure(self, dir_, get, top):
        """Add configured blocks to provided dir (space separated config option is auto-magically received via provided getter)."""
        dir_.add_dir(self, name_filter=self.validate(get(self.extra_option(top=top), "")))


#: Quiet, non-interactive, least invasive and loggable apt-get call (Dpkg::Use-Pty=false is to avoid https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=539617).
APT_GET = "apt-get --quiet --yes --option=APT::Install-Recommends=false --option=Acquire::Languages=none --option=Dpkg::Use-Pty=false"

CONFIG_BLOCKS = ConfigBlocks()

CONFIG_BLOCKS.add(mini_buildd.files.PerlModule("ccache",
                                               description="Enable ccache",
                                               snippet="""\
$path = "/usr/lib/ccache:$path";
$$build_environment{'CCACHE_DIR'} = '%LIBDIR%/.ccache';
"""))

SETUP_BLOCKS = SetupBlocks()

SETUP_BLOCKS.add(mini_buildd.files.ShScript("ccache",
                                            description="Have 'ccache' ready in builds",
                                            snippet=f"""\
{APT_GET} install ccache
"""))

SETUP_BLOCKS.add(mini_buildd.files.ShScript("eatmydata",
                                            description="Install 'eatmydata' when available",
                                            snippet=f"""\
if {APT_GET} install eatmydata; then
    # eatmydata <= 26: Just one 'deb', and not multiarched
    EATMYDATA_SO=$(dpkg -L eatmydata | grep 'libeatmydata.so$') || true
    if [ -z "${{EATMYDATA_SO}}" ]; then
        # eatmydata >= 82: Has a library 'deb', and is multiarched
        EATMYDATA_SO=$(dpkg -L libeatmydata1 | grep 'libeatmydata.so$') || true
    fi
    if [ -n "${{EATMYDATA_SO}}" ]; then
        printf " ${{EATMYDATA_SO}}" >> /etc/ld.so.preload
    else
        printf "W: eatmydata: No *.so found (skipping)\\n" >&2
    fi
else
    printf "W: eatmydata: Not installable (skipping)\\n" >&2
fi
"""))

SETUP_BLOCKS.add(mini_buildd.files.BashScript("disable-check-valid-until",
                                              description="Disable APT's 'Check-Valid-Until' in old releases that really need it",
                                              snippet="""\
# Pre 1.99.19, this used to live here: /etc/schroot/setup.d/15mini-buildd-workarounds

# (Some) archived dists have a fixed Valid-Util header.
#
# We still want these to work, however.
#
# This can (and should) be configured per source, however dists older
# stretch won't honor the per-source '[check-valid-until=no]' apt
# option, and we *must* do it per global apt config.
#
# Dists >= stretch will never need this, thus there should be no need
# to update this workaround ever again.
case ${SCHROOT_CHROOT_NAME} in
  *-etch-*|*-lenny-*|*-squeeze-*|*-wheezy-*|*-jessie-*)
    printf "W: Disabling 'Check-Valid-Until' globally in ${SCHROOT_CHROOT_NAME}: Archived, and so old it does not support [check-valid-until=no] in the apt line.\\n"
    printf "Acquire::Check-Valid-Until \"false\";\\n" >"/etc/apt/apt.conf.d/10disable-check-valid-until"
    ;;
esac
"""))

SETUP_BLOCKS.add(mini_buildd.files.ShScript("accept-sun-java6-license",
                                            description="Accept sun-java6 licence (debconf) so we can build-depend on it",
                                            snippet="""\
echo "sun-java6-bin shared/accepted-sun-dlj-v1-1 boolean true" | debconf-set-selections --verbose
echo "sun-java6-jdk shared/accepted-sun-dlj-v1-1 boolean true" | debconf-set-selections --verbose
echo "sun-java6-jre shared/accepted-sun-dlj-v1-1 boolean true" | debconf-set-selections --verbose
"""))

SETUP_BLOCKS.add(mini_buildd.files.ShScript("bash-building-debbug327477",
                                            description="Fixup /dev/fd|stdin|stdout|stderr (workaround for Debian Bug #327477 (bash building))",
                                            snippet="""\
[ -e /dev/fd ] || ln -sv /proc/self/fd /dev/fd
[ -e /dev/stdin ] || ln -sv fd/0 /dev/stdin
[ -e /dev/stdout ] || ln -sv fd/1 /dev/stdout
[ -e /dev/stderr ] || ln -sv fd/2 /dev/stderr
"""))

SETUP_BLOCKS.add(mini_buildd.files.ShScript("experimental-squeeze-apt-fix",
                                            description="Try some hard configs for very urold releases to make APT work",
                                            snippet="""\
rm -v -r -f /var/lib/apt/lists/*
APT_FIX="/etc/apt/apt.conf.d/mini-buildd-squeeze-apt-fix"
printf "%s\\n" "Acquire::http::No-Cache true;" "Acquire::http::Pipeline-Depth 0;" "APT::Get::Force-Yes true;" >"${APT_FIX}"
cat "${APT_FIX}"
"""))

SETUP_BLOCKS.add(mini_buildd.files.BashScript("obsolete-debbug728096-schroot-shm-fix",
                                              description="Fix /dev/shm mount (workaround for Debian Bug #728096)",
                                              snippet="""\
# Pre 1.99.19, this used to live here: /etc/schroot/setup.d/15mini-buildd-workarounds

# '/dev/shm' might be a symlink to '/run/shm' in some
# chroots. Depending on the system mini-buildd/sbuild runs on
# and what chroot is being build for this may or may not lead to
# one or both of these problems:
#
# - shm not mounted properly in build chroot (leading to build errors when shm is used).
# - shm mount on *the host* gets overloaded (leading to an shm mount "leaking" on the host).
#
# This workaround
#
# - [in '/etc/schroot/mini-buildd/fstab-generic'] Removes shm form the generic sbuild mounts (avoids the mount leaking).
# - [here] Fixes /dev/shm to a directory in case it's not.
# - [here] Always mounts /dev/shm itself.
#
# Debian Bug: https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=728096
#
# When this is fixed (and we have added a versioned depends on
# schroot accordingly), all of the above may be reverted again.
#
printf "=> Fixing up /dev/shm mount (Bug #728096):\\n"

if [ -L "/dev/shm" ]; then
  printf "Removing /dev/shm symlink...\\n"
  rm -v "/dev/shm"
fi
mkdir -v -p "/dev/shm"
mount -v -ttmpfs none "/dev/shm"
"""))
