import http
import logging
import json
import glob

import django.http
import django.template.response
import django.views

import mini_buildd.api

LOG = logging.getLogger(__name__)


def context(_request):
    return {"mbd": {"version": mini_buildd.__version__,
                    "config": mini_buildd.config,
                    "api": mini_buildd.api,
                    "daemon": mini_buildd.Daemon()}}


class ExceptionMiddleware():
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)

    @classmethod
    def __error(cls, request, exception, status=http.HTTPStatus.INTERNAL_SERVER_ERROR):
        http_exception = mini_buildd.e2http(exception, status)
        return django.template.response.TemplateResponse(request,
                                                         "error.html",
                                                         http_exception.rfc7807.to_json(),
                                                         status=http_exception.status)

    @classmethod
    def process_exception(cls, request, exception):
        return cls.__error(request, exception)

    # Despite having a custom middleware to handle exceptions, the following django default functions still need to be overwritten (see also: urls.py).
    @classmethod
    def bad_request(cls, request, exception):
        return cls.__error(request, exception, http.HTTPStatus.BAD_REQUEST)

    @classmethod
    def permission_denied(cls, request, exception):
        return cls.__error(request, exception, http.HTTPStatus.UNAUTHORIZED)

    @classmethod
    def page_not_found(cls, request, exception):
        return cls.__error(request, exception, http.HTTPStatus.NOT_FOUND)

    @classmethod
    def server_error(cls, request):
        return cls.__error(request, None)


class PathView(django.views.generic.base.TemplateView):
    """Compute template name from request path."""

    def get_template_names(self):
        return self.request.path.lstrip("/")


class BuildlogView(django.views.generic.base.TemplateView):
    template_name = "mini_buildd/buildlog.html"

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)

        file_name = kwargs["path"] + ".buildlog"
        resource = {"live": mini_buildd.config.HOME["spool"], "log": mini_buildd.config.HOME["events"]}[kwargs["type"]]
        ctx.update({"file_path": resource.path.get(file_name),
                    "download_uri": resource.uris["static"].get(file_name) if resource.uris else None})
        return ctx


class EventsLogView(django.views.generic.base.TemplateView):
    template_name = "mini_buildd/events-log/log.html"

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx.update({
            "log": mini_buildd.changes.Log(kwargs["distribution"], kwargs["source"], kwargs["version"]),
        })
        return ctx


class EventsLogHomeView(django.views.generic.base.TemplateView):
    template_name = "mini_buildd/events-log/home.html"

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)

        events = []
        for j in glob.glob(mini_buildd.config.HOME["events"].path.get("*", "*", "*.json")):
            with open(j) as f:
                events.append(mini_buildd.events.Event.from_json(json.load(f)))

        ctx.update({"events": events})
        return ctx


class API(django.views.generic.base.TemplateView):
    def dispatch(self, request, *args, **kwargs):  # pylint: disable=too-many-return-statements,too-many-branches
        self.mbd_api_cmd = None  # pylint: disable=attribute-defined-outside-init
        try:
            if request.method != 'GET':
                raise mini_buildd.HTTPBadRequest("API allows GET requests only")

            # Get API class from 'command' parameter
            command = request.GET.get("command", None)
            api_cls = mini_buildd.api.COMMANDS.get(command, None)
            if api_cls is None or api_cls.isgroup():
                raise mini_buildd.HTTPBadRequest(f"Unknown command: {command}")

            # Authentication
            auth_err = api_cls.auth_err(request.user)
            if auth_err:
                raise mini_buildd.HTTPUnauthorized(auth_err)

            # Generate command object
            self.mbd_api_cmd = api_cls(request.GET, request=request)  # pylint: disable=attribute-defined-outside-init

            # Check if we need a running daemon
            if api_cls.NEEDS_RUNNING_DAEMON and not mini_buildd.Daemon().is_alive():
                raise mini_buildd.HTTPUnavailable(f"Command needs running daemon: {command}")

            # Check confirmable calls
            if api_cls.CONFIRM and request.GET.get("confirm", None) != command:
                raise mini_buildd.HTTPUnauthorized(f"Command needs to be confirmed: {command}")

            # Show api command name and user calling it.
            LOG.info(f"API call '{command}' by user '{request.user}'")

            # Run API call (dep-injection via daemon object)
            self.mbd_api_cmd.run()

            # Generate API call output
            return django.http.JsonResponse(self.mbd_api_cmd.result)

        except Exception as e:
            http_exception = mini_buildd.e2http(e)
            return django.http.JsonResponse(http_exception.rfc7807.to_json(),
                                            status=http_exception.status,
                                            content_type="application/problem+json")


class APIView(API):
    def dispatch(self, request, *args, **kwargs):  # pylint: disable=too-many-return-statements,too-many-branches
        result = json.loads(super().dispatch(request, *args, **kwargs).getvalue())
        if result.get("status", 200) != 200:
            raise mini_buildd.HTTPError(result["status"], result["detail"])
        return django.template.response.TemplateResponse(request,
                                                         [mini_buildd.config.URIS['api']['view'].get(f"{self.mbd_api_cmd.name()}.html"),
                                                          mini_buildd.config.URIS['api']['view'].get("default.html")],
                                                         {**self.get_context_data(),
                                                          **{"api_cmd": self.mbd_api_cmd,
                                                             "repositories": mini_buildd.models.repository.Repository.mbd_get_prepared()}})


class AccountProfileView(django.views.generic.base.TemplateView):
    template_name = "mini_buildd/account_profile.html"
